#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_rename_column_headers import \
    BULK_OLD_RENAME_STEP


def upgrade_simple_import_1_to_2_and_rename(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds a rename step after a simple import, so that it is consistent
    with how simple import used to rename column headers after import.

    Also makes sure the old id algorithms is used to create the step.

    Old format: {
        "step_version": 1, 
        "step_type": "simple_import", 
        "params": {
            file_names: ['data.csv'],
        }
    }

    New format: [
        {
            "step_version": 2, 
            "step_type": "simple_import", 
            "params": {
                file_names: ['data.csv'],
                use_deprecated_id_algorithm: true
            }
        },
        {
            "step_version": 1, 
            "step_type": "bulk_old_rename", 
            "params": {}
        },
    ]    
    """
    step['step_version'] = 2
    step['params']['use_deprecated_id_algorithm'] = True

    return [step, BULK_OLD_RENAME_STEP] + later_steps



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from typing import Any, Dict, List, Optional, Tuple, Union

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.code_chunk_utils import \
    get_right_combine_with_column_delete_code_chunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import \
    DeleteColumnsCodeChunk
from mitosheet.parser import parse_formula
from mitosheet.state import State
from mitosheet.types import (FORMULA_ENTIRE_COLUMN_TYPE, ColumnID,
                             FormulaAppliedToType)
from mitosheet.transpiler.transpile_utils import column_header_list_to_transpiled_code


class SetColumnFormulaCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, formula_label: Union[str, bool, int, float], index_labels_formula_is_applied_to: FormulaAppliedToType, new_formula: str, public_interface_version: int):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_id = column_id
        self.formula_label = formula_label
        self.index_labels_formula_is_applied_to = index_labels_formula_is_applied_to
        self.new_formula = new_formula
        self.public_interface_version = public_interface_version

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Updated column formula'
    
    def get_description_comment(self) -> str:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        if self.index_labels_formula_is_applied_to['type'] == FORMULA_ENTIRE_COLUMN_TYPE:
            return f'Set formula of {column_header}'
        else:
            return f'Set formula of {column_header} at rows { {column_header_list_to_transpiled_code(self.index_labels_formula_is_applied_to["index_labels"])}}' # type: ignore

    def get_code(self) -> Tuple[List[str], List[str]]:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        python_code, _, _, _ = parse_formula(
            self.new_formula, 
            column_header,
            self.formula_label,
            self.index_labels_formula_is_applied_to,
            self.post_state.dfs[self.sheet_index],
            df_name=self.post_state.df_names[self.sheet_index],
        )

        return [
            python_code
        ], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_column_code_chunk(self, other_code_chunk: DeleteColumnsCodeChunk) -> Optional[CodeChunk]:
        return get_right_combine_with_column_delete_code_chunk(
            self,
            other_code_chunk,
            'sheet_index',
            'column_id',
        )

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_column_code_chunk(other_code_chunk)

        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Preprocessing occurs to data that is imported into the sheet, and occurs
to _any_ data that is input into the mitosheet - they always occur, and
require no user input. 
"""

from typing import List, Type
from mitosheet.preprocessing.preprocess_step_performer import PreprocessStepPerformer
from mitosheet.preprocessing.preprocess_read_file_paths import ReadFilePathsPreprocessStepPerformer
from mitosheet.preprocessing.preprocess_check_args_type import CheckArgsTypePreprocessStepPerformer
from mitosheet.preprocessing.preprocess_copy import CopyPreprocessStepPerformer


# NOTE: These should be in the order you want to apply them to the arguments,
# as they are run in a linear order
PREPROCESS_STEP_PERFORMERS: List[Type[PreprocessStepPerformer]] = [
   # First, we make sure all the args are the right type
   CheckArgsTypePreprocessStepPerformer,
   # Then, we copy the args to make sure we don't change them accidently
   CopyPreprocessStepPerformer,
   # Then, we read in the files
   ReadFilePathsPreprocessStepPerformer,
]




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
For going to a string series.
"""

import pandas as pd
import numpy as np
from typing import Any

from mitosheet.is_type_utils import is_bool_dtype, is_datetime_dtype, is_number_dtype, is_string_dtype, is_timedelta_dtype


def to_string_series(
        unknown_object: Any,
        on_uncastable_arg_element: Any=('default', np.NaN), #Union[Literal['error'], Tuple[Literal['default'], any]]
    ) -> pd.Series:
    """
    Converts the given object to a string series. Note that on_uncastable_arg_element
    is irrelevant here, as anything can be turned into a string!
    """

    # If it is not a series, we put it in a series, and set it to a string
    if not isinstance(unknown_object, pd.Series):
        return pd.Series([str(unknown_object)], dtype='str')

    column_dtype = str(unknown_object.dtype)
    if is_bool_dtype(column_dtype):
        return unknown_object.astype('str')
    elif is_datetime_dtype(column_dtype):
        return unknown_object.dt.strftime('%Y-%m-%d %X')
    elif is_timedelta_dtype(column_dtype):
        return unknown_object.astype('str')
    elif is_number_dtype(column_dtype):
        return unknown_object.astype('str')
    elif is_string_dtype(column_dtype):
        # We need to cast here, because object series are treated
        # as string series, and may contain other types
        return unknown_object.astype('str')
    else:
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that are useful for control flow. For now, this
is just IF statements.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""

from typing import Optional
import pandas as pd

from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.sheet_functions.utils import get_final_result_series_or_primitive, get_series_from_primitive_or_series
from mitosheet.public.v3.types.decorators import cast_values_in_all_args_to_type, cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import AnyPrimitiveOrSeriesInputType, BoolFunctionReturnType, BoolInputType, BoolRestrictedInputType


@cast_values_in_all_args_to_type('bool')
@handle_sheet_function_errors
def AND(*argv: Optional[BoolInputType]) -> BoolFunctionReturnType:
    """
    {
        "function": "AND",
        "description": "Returns True if all of the provided arguments are True, and False if any of the provided arguments are False.",
        "search_terms": ["and", "&", "if", "conditional"],
        "examples": [
            "AND(True, False)",
            "AND(Nums > 100, Nums < 200)",
            "AND(Pay > 10, Pay < 20, Status == 'active')"
        ],
        "syntax": "AND(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    # If we don't find any arguements, we default to True. Excel in practice
    # defaults to a value error if there are no args
    return get_final_result_series_or_primitive(
        True,
        argv,
        lambda df: df.all().all(),
        lambda previous_value, new_value: previous_value and new_value,
        lambda previous_series, new_series: previous_series & new_series
    )

@cast_values_in_arg_to_type('series', 'bool')
@handle_sheet_function_errors
def BOOL(series: BoolRestrictedInputType) -> BoolFunctionReturnType:
    """
    {
        "function": "BOOL",
        "description": "Converts the passed arguments to boolean values, either True or False. For numberic values, 0 converts to False while all other values convert to True.",
        "search_terms": ["bool", "boolean", "true", "false", "dtype", "convert"],
        "examples": [
            "BOOL(Amount_Payed)",
            "AND(BOOL(Amount_Payed), Is_Paying)"
        ],
        "syntax": "BOOL(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "An series to convert to boolean values, either True or False."
            }
        ]
    }
    """
    if isinstance(series, bool):
        return series

    return series.fillna(False)

@cast_values_in_arg_to_type('condition', 'bool')
@handle_sheet_function_errors
def IF(condition: pd.Series, true_series: AnyPrimitiveOrSeriesInputType, false_series: AnyPrimitiveOrSeriesInputType) -> pd.Series:
    """
    {
        "function": "IF",
        "description": "Returns one value if the condition is True. Returns the other value if the conditon is False.",
        "search_terms": ["if", "conditional", "and", "or"],
        "examples": [
            "IF(Status == 'success', 1, 0)",
            "IF(Nums > 100, 100, Nums)",
            "IF(AND(Grade >= .6, Status == 'active'), 'pass', 'fail')"
        ],
        "syntax": "IF(boolean_condition, value_if_true, value_if_false)",
        "syntax_elements": [{
                "element": "boolean_condition",
                "description": "An expression or series that returns True or False values. Valid conditions for comparison include ==, !=, >, <, >=, <=."
            },
            {
                "element": "value_if_true",
                "description": "The value the function returns if condition is True."
            },
            {
                "element": "value_if_false",
                "description": "The value the function returns if condition is False."
            }
        ]
    }
    """

    true_series = get_series_from_primitive_or_series(true_series, condition.index)
    false_series = get_series_from_primitive_or_series(false_series, condition.index)

    return pd.Series(
        data=[true_series.loc[i] if c else false_series.loc[i] for i, c in condition.items()],
        index=condition.index
    )


@cast_values_in_all_args_to_type('bool')
@handle_sheet_function_errors
def OR(*argv: Optional[BoolInputType]) -> BoolFunctionReturnType:
    """
    {
        "function": "OR",
        "description": "Returns True if any of the provided arguments are True, and False if all of the provided arguments are False.",
        "search_terms": ["or", "if", "conditional"],
        "examples": [
            "OR(True, False)",
            "OR(Status == 'success', Status == 'pass', Status == 'passed')"
        ],
        "syntax": "OR(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    # If we don't find any arguements, we default to True. Excel in practice
    # defaults to a value error if there are no args
    return get_final_result_series_or_primitive(
        False,
        argv,
        lambda df: df.any().any(),
        lambda previous_value, new_value: previous_value or new_value,
        lambda previous_series, new_series: previous_series | new_series
    )


# TODO: we should see if we can list these automatically!
CONTROL_FUNCTIONS = {
    'AND': AND,
    'BOOL': BOOL,
    'IF': IF,
    'OR': OR,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List


def upgrade_snowflake_import_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Changes parameter from table to table_or_view


    OLD: 
    {
        "step_version": 1, 
        "step_type": "snowflake_import", 
        "params": {
            "table_loc_and_warehouse": {
                "warehouse": str,
                "database": str,
                "schema": str,
                "table": str
            },
            "query_params": {
                "columns": str[],
                "limit": int | undefined 
            },
        }
    }

    NEW: 
    {
        "step_version": 2, 
        "step_type": "snowflake_import", 
        "params": {
            "table_loc_and_warehouse": {
                "warehouse": str,
                "database": str,
                "schema": str,
                "table_or_view": str
            },
            "query_params": {
                "columns": str[],
                "limit": int | undefined 
            },
        }
    }
    """

    params = step['params']
    params['table_loc_and_warehouse']['table_or_view'] = params['table_loc_and_warehouse']['table']
    del params['table_loc_and_warehouse']['table']


    return [{
        "step_version": 2, 
        "step_type": "snowflake_import", 
        "params": params
    }] + later_steps


def upgrade_snowflake_import_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Adds "role": None to the table_loc_and_warehouse so previous analyses continue to use the default role

    OLD: 
    {
        "step_version": 1, 
        "step_type": "snowflake_import", 
        "params": {
            "table_loc_and_warehouse": {
                "warehouse": str,
                "database": str,
                "schema": str,
                "table": str
            },
            "query_params": {
                "columns": str[],
                "limit": int | undefined 
            },
        }
    }

    NEW: 
    {
        "step_version": 2, 
        "step_type": "snowflake_import", 
        "params": {
            "table_loc_and_warehouse": {
                "role": None
                "warehouse": str,
                "database": str,
                "schema": str,
                "table_or_view": str
            },
            "query_params": {
                "columns": str[],
                "limit": int | undefined 
            },
        }
    }
    """

    params = step['params']
    params['table_loc_and_warehouse']['role'] = None

    return [{
        "step_version": 3, 
        "step_type": "snowflake_import", 
        "params": params
    }] + later_steps



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.drop_duplicates_code_chunk import DropDuplicatesCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID

class DropDuplicatesStepPerformer(StepPerformer):
    """
    Allows you to drop duplicates from a dataframe
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'drop_duplicates'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index = get_param(params, 'sheet_index')
        column_ids = get_param(params, 'column_ids')
        keep = get_param(params, 'keep')

        column_headers = [
            prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
            for column_id in column_ids
        ]

        # If the subset is none, then we don't actually do the drop, as there are no
        # duplicates between 0 columns
        if len(column_headers) == 0:
            return prev_state, {}

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        pandas_start_time = perf_counter()
        final_df = post_state.dfs[sheet_index].drop_duplicates(
            subset=column_headers,
            keep=keep
        )
        pandas_processing_time = perf_counter() - pandas_start_time


        post_state.dfs[sheet_index] = final_df

        # We calculate the number of rows dropped, so we can return this to the frontend
        num_rows_dropped = len(prev_state.dfs[sheet_index].index) - len(post_state.dfs[sheet_index].index)

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'result': {
                'num_rows_dropped': num_rows_dropped
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DropDuplicatesCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'keep')
            )   
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.concat_code_chunk import ConcatCodeChunk
from mitosheet.state import DATAFRAME_SOURCE_CONCAT, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param


class ConcatStepPerformer(StepPerformer):
    """
    Allows you to concatenate two or more dataframes together.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'concat'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        join = get_param(params, 'join') # inner | outter
        ignore_index: bool = get_param(params, 'ignore_index')
        sheet_indexes: List[int] = get_param(params, 'sheet_indexes')

        post_state = prev_state.copy()

        to_concat = [post_state.dfs[sheet_index] for sheet_index in sheet_indexes]

        if len(to_concat) == 0:
            new_df = pd.DataFrame()
            pandas_processing_time = 0.0
        else:
            pandas_start_time = perf_counter()
            new_df = pd.concat(to_concat, join=join, ignore_index=ignore_index)
            pandas_processing_time = perf_counter() - pandas_start_time

        # Add this dataframe to the new post state
        post_state.add_df_to_state(new_df, DATAFRAME_SOURCE_CONCAT)

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ConcatCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'join'),
                get_param(params, 'ignore_index'),
                get_param(params, 'sheet_indexes'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk
from mitosheet.code_chunks.step_performers.dataframe_steps.dataframe_rename_code_chunk import DataframeRenameCodeChunk
from mitosheet.state import State

class DataframeDeleteCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_indexes: List[int], old_dataframe_names: List[str]):
        super().__init__(prev_state, post_state)
        self.sheet_indexes = sheet_indexes
        self.old_dataframe_names = old_dataframe_names

    def get_display_name(self) -> str:
        return 'Deleted Dataframe'
    
    def get_description_comment(self) -> str:
        return f'Deleted {", ".join(self.old_dataframe_names)}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        # NOTE: We do not actually delete the dataframes, as it serves no purpose, but 
        # makes things confusing if you're importing dataframes from outside the sheet
        # However, we still use this code chunk as it optimizes out other code chunks
        return [], []

    def _combine_right_dataframe_delete(self, other_code_chunk: "DataframeDeleteCodeChunk") -> CodeChunk:
        first_sheet_indexes = deepcopy(self.sheet_indexes)
        second_sheet_indexes = deepcopy(other_code_chunk.sheet_indexes)

        # Because we don't have sheet ids, we need to bump any deleted dataframes
        # that are greater than those deleted first, so that they have the correct
        # index in the newly combined step
        for first_sheet_index in first_sheet_indexes:
            for index, second_sheet_index in enumerate(second_sheet_indexes):
                if first_sheet_index <= second_sheet_index:
                    second_sheet_indexes[index] = second_sheet_index + 1

        sheet_indexes = first_sheet_indexes + second_sheet_indexes
        old_dataframe_names = self.old_dataframe_names + other_code_chunk.old_dataframe_names

        return DataframeDeleteCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            sheet_indexes,
            old_dataframe_names
        )


    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, DataframeDeleteCodeChunk):
            return self._combine_right_dataframe_delete(other_code_chunk)
        
        return None

    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:

        # Dataframe renames are compicated to combine with dataframe deletes so we simple
        # do not combine them
        if isinstance(other_code_chunk, DataframeRenameCodeChunk):
            return None

        deleted_sheet_indexes = self.sheet_indexes
        created_sheet_indexes = other_code_chunk.get_created_sheet_indexes()


        if created_sheet_indexes is not None and set(deleted_sheet_indexes) == set(created_sheet_indexes):
            # If all we did was create the dfs we deleted, we can just return a no op
            return NoOpCodeChunk(other_code_chunk.prev_state, self.post_state)
            
        elif created_sheet_indexes is not None and set(deleted_sheet_indexes).issuperset(set(created_sheet_indexes)):
            # If the set we are deleting is a superset of the the dataframes that we created
            # in this other step, then we must create a new code chunk that doesn't have these
            # deletes for the created sheets (as we're just not including these created steps)
            new_deleted_sheet_indexes = copy(self.sheet_indexes)
            new_old_dataframe_names = copy(self.old_dataframe_names)

            for created_sheet_index in created_sheet_indexes:
                # Find the index of the sheet index we need to delete, and remove that as well
                # as the old dataframe name, which is at the same index
                index_in_params_array = new_deleted_sheet_indexes.index(created_sheet_index)
                new_deleted_sheet_indexes.pop(index_in_params_array)
                new_old_dataframe_names.pop(index_in_params_array)
            
            return DataframeDeleteCodeChunk(
                other_code_chunk.prev_state,
                self.post_state,
                new_deleted_sheet_indexes,
                new_old_dataframe_names,
            )

        sheet_indexes_edited = other_code_chunk.get_edited_sheet_indexes()
        if sheet_indexes_edited is not None and set(deleted_sheet_indexes).issuperset(set(sheet_indexes_edited)):
            return self

        return None





from collections import OrderedDict
import hashlib
import json
import os
import pickle
from typing import Any, Dict, List, Callable, Optional, Tuple, Union

import pandas as pd

from mitosheet.mito_backend import MitoBackend
from mitosheet.types import CodeOptions
from mitosheet.utils import get_new_id

def _get_dataframe_hash(df: pd.DataFrame) -> bytes:
    """
    Returns a hash for a pandas dataframe that is consistent across runs, notably including:
    1. The column names
    2. The values of the dataframe
    3. The index of the dataframe
    4. The order of all of these

    This is necessary due to the issues described here: https://github.com/streamlit/streamlit/issues/7086
    where streamlit default hashing is not ideal for pandas dataframes, as it misses some column header and
    reordering changes. 
    """
    try:
        return hashlib.md5(
            bytes(str(pd.util.hash_pandas_object(df.columns)), 'utf-8') +
            bytes(str(pd.util.hash_pandas_object(df)), 'utf-8')
        ).digest()
    except TypeError as e:        
        # Use pickle if pandas cannot hash the object for example if
        # it contains unhashable objects.
        return b"%s" % pickle.dumps(df, pickle.HIGHEST_PROTOCOL)

def get_dataframe_hash(df: pd.DataFrame) -> bytes:
    _PANDAS_ROWS_LARGE = 100000
    _PANDAS_SAMPLE_SIZE = 10000
    
    if len(df) >= _PANDAS_ROWS_LARGE:
        df = df.sample(n=_PANDAS_SAMPLE_SIZE, random_state=0)
    
    return _get_dataframe_hash(df)

try:
    import streamlit.components.v1 as components
    import streamlit as st

    parent_dir = os.path.dirname(os.path.abspath(__file__))

    mito_build_dir = os.path.join(parent_dir, "mitoBuild")
    _mito_component_func = components.declare_component("my_component", path=mito_build_dir)

    message_passer_build_dr = os.path.join(parent_dir, "messagingBuild")
    _message_passer_component_func = components.declare_component("message-passer", path=message_passer_build_dr)


    def get_session_id() -> Optional[str]:
        """
        This returns the session id for the current script run. Notably, this is different
        when the user:
        1. Refreshes the page
        2. Closes the page and reopens it
        3. Is a different user
        4. Is a different browser

        It allows us to cache the Mito backend on the session id, so that it is not
        shared across users. Notably, it clearing when refreshed is the same behavior
        that streamlit caching has by default -- e.g. the button component will reset
        when the page is refreshed.

        From the streamlit docs:
        A context object that contains data for a "script run" - that is,
        data that's scoped to a single ScriptRunner execution (and therefore also
        scoped to a single connected "session").
        """
        from streamlit.runtime.scriptrunner import get_script_run_ctx
        ctx = get_script_run_ctx()
        if ctx is None:
            return None
        return ctx.session_id

    @st.cache_resource(hash_funcs={pd.DataFrame: get_dataframe_hash})
    def _get_mito_backend(
            *args: Union[pd.DataFrame, str, None], 
            _importers: Optional[List[Callable]]=None, 
            _sheet_functions: Optional[List[Callable]]=None, 
            _code_options: Optional[CodeOptions]=None,
            import_folder: Optional[str]=None,
            df_names: Optional[List[str]]=None,
            session_id: Optional[str]=None,
            key: Optional[str]=None # So it caches on key
        ) -> Tuple[MitoBackend, List[Any]]: 

        mito_backend = MitoBackend(
            *args, 
            import_folder=import_folder,
            user_defined_importers=_importers, user_defined_functions=_sheet_functions,
            code_options=_code_options,
        )

        # Make a send function that stores the responses in a list
        responses = []
        def send(response):
            responses.append(response)
        
        mito_backend.mito_send = send

        if df_names is not None and len(df_names) > 0:
            mito_backend.receive_message(
                {
                    'event': 'update_event',
                    'id': get_new_id(),
                    'type': 'args_update',
                    'params': {
                        'args': df_names
                    },
                }
            )

        return mito_backend, responses

    def message_passer_component(key: Optional[str]=None) -> Any:
        """
        This component simply passes messages from the frontend to the backend,
        so that the backend can process them before it is rendered.
        """
        component_value = _message_passer_component_func(key=key)
        return component_value


    def spreadsheet( # type: ignore
            *args: Union[pd.DataFrame, str, None], 
            sheet_functions: Optional[List[Callable]]=None, 
            importers: Optional[List[Callable]]=None, 
            df_names: Optional[List[str]]=None,
            import_folder: Optional[str]=None,
            code_options: Optional[CodeOptions]=None,
            key=None
        ) -> Tuple[OrderedDict[str, pd.DataFrame], str]:
        """
        Create a new instance of the Mito spreadsheet in a streamlit app.

        Parameters
        ----------
        args: pd.Dataframe or str or None
            The arguments to pass to the Mito spreadsheet. If a dataframe is
            passed, it will be displayed as a sheet tab. If a string is passed,
            it will be read in with a pd.read_csv call. If None is passed, it 
            will be skipped.
        sheet_functions: List[Callable]
            A list of functions that can be used in the spreadsheet. Functions
            should be capitalized.
        importers: List[Callable]
            A list of functions that can be used to import dataframes. Each
            function should return a dataframe. 
        df_names: List[str]
            A list of names for the dataframes passed in. If None, the dataframes
            will be named df0, df1, etc.
        key: str or None
            An key that uniquely identifies this component. This must be passed
            for now, or the component will not work. Not sure why.

        Returns
        -------
        Tuple[Dict[str, pd.DataFrame], List[str]]
            A tuple. The first element is a mapping from dataframe names to the
            final dataframes. The second element is a list of lines of code
            that were executed in the Mito spreadsheet.
        """
        # Get the absolute path to the import_folder, in case it is relative. Also
        # check that this folder exists, and throw an error if it does not.
        if import_folder is not None:
            import_folder = os.path.expanduser(import_folder)
            import_folder = os.path.abspath(import_folder)

            if not os.path.exists(import_folder):
                raise ValueError(f"Import folder {import_folder} does not exist. Please change the file path or create the folder.")

        session_id = get_session_id()

        mito_backend, responses = _get_mito_backend(
            *args, 
            _sheet_functions=sheet_functions,
            _importers=importers, 
            _code_options=code_options,
            import_folder=import_folder,
            session_id=session_id,
            df_names=df_names, 
            key=key
        )

        # Mito widgets need new ids every time a new one is displayed. As such, if
        # the key is None, we generate a new one. Notably, we do this after getting the
        # mito_backend, so that we can cache the mito_backend on the user provided key.
        if key is None:
            key = mito_backend.analysis_name

        sheet_data_json = mito_backend.steps_manager.sheet_data_json,
        analysis_data_json = mito_backend.steps_manager.analysis_data_json,
        user_profile_json = mito_backend.get_user_profile_json()

        msg = message_passer_component(key=str(key) + 'message_passer')
        if (
            msg is not None \
            and msg['id'] not in [response['id'] for response in responses] \
            and msg['analysis_name'] == mito_backend.analysis_name
        ):
            # We receive a message if:
            # 1. It is not None
            # 2. We have not already received it on this backend
            # 3. It is for this analysis. 
            # Note that the final two conditions are to prevent messages that have been sent
            # by the message passer component from being received again. This happens because
            # when a component value is set, it is always returned by the message_passer_component
            # until a new component value is set            
            mito_backend.receive_message(msg)
            
        responses_json = json.dumps(responses)

        _mito_component_func(
            key=key, 
            sheet_data_json=sheet_data_json, analysis_data_json=analysis_data_json, user_profile_json=user_profile_json, 
            responses_json=responses_json, id=id(mito_backend)
        )

        # We return a mapping from dataframe names to dataframes
        final_state = mito_backend.steps_manager.curr_step.final_defined_state
        code = mito_backend.steps_manager.code()

        ordered_dict = OrderedDict()
        for df_name, df in zip(final_state.df_names, final_state.dfs):
            ordered_dict[df_name] = df

        return ordered_dict, "\n".join(code)
    
except ImportError:
    def spreadsheet(*args, key=None): # type: ignore
        raise RuntimeError("Couldn't import streamlit. Install streamlit with `pip install streamlit` to use the mitosheet component.")






from datetime import datetime, timedelta
from typing import Type, Union

import pandas as pd

from mitosheet.public.v3.rolling_range import RollingRange

# Input types
StringRestrictedInputType = Union[pd.Series, str]
StringInputType = Union[pd.DataFrame, RollingRange, StringRestrictedInputType]

IntRestrictedInputType = Union[pd.Series, int]
IntInputType = Union[pd.DataFrame, RollingRange, IntRestrictedInputType]

FloatRestrictedInputType = Union[pd.Series, float]
FloatInputType = Union[pd.DataFrame, RollingRange, FloatRestrictedInputType]

NumberRestrictedInputType = Union[IntRestrictedInputType, FloatRestrictedInputType]
NumberInputType = Union[IntInputType, FloatInputType]

BoolRestrictedInputType = Union[pd.Series, bool]
BoolInputType = Union[pd.DataFrame, RollingRange, BoolRestrictedInputType]

DatetimeRestrictedInputType = Union[pd.Series, datetime, pd.Timestamp]
DatetimeInputType = Union[pd.DataFrame, RollingRange, DatetimeRestrictedInputType]

AnyPrimitiveInputType = Union[str, int, float, bool, datetime, timedelta]
AnyPrimitiveOrSeriesInputType = Union[AnyPrimitiveInputType, pd.Series]


# Return types
StringFunctionReturnType = Union[pd.Series, str]
IntFunctionReturnType = Union[pd.Series, int]
FloatFunctonReturnType = Union[pd.Series, float]
NumberFunctionReturnType = Union[pd.Series, int, float]
BoolFunctionReturnType = Union[pd.Series, bool]
DatetimeFunctionReturnType = Union[pd.Series, datetime]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict

from mitosheet.errors import MitoError
from mitosheet.step import Step
from mitosheet.step_performers.import_steps.snowflake_import import SnowflakeImportStepPerformer
from mitosheet.types import StepsManagerType
from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer
from mitosheet.step_performers.import_steps.excel_import import ExcelImportStepPerformer
from mitosheet.step_performers.import_steps.excel_range_import import ExcelRangeImportStepPerformer

CSV_IMPORT_ERROR = 'There was an error importing this CSV file. Make sure that the file exists, or select a different file or dataframe.'
EXCEL_IMPORT_ERROR = 'There was an error importing this sheet. Make sure that the Excel file exists and contains this sheet, or select a different file or dataframe.'
DATAFRAME_IMPORT_ERROR = 'There was an error importing this dataframe. Make sure that the dataframe is defined, or select a different file or dataframe.'
SNOWFLAKE_IMPORT_ERROR = 'There was an error executing this query. Make sure that your credentials are valid and you have access to this table, or select a different file or dataframe.'


def get_import_error_for_step_type(step_type: str) -> str:
    if step_type == SimpleImportStepPerformer.step_type():
        return CSV_IMPORT_ERROR
    if step_type == ExcelImportStepPerformer.step_type() or step_type == ExcelRangeImportStepPerformer.step_type():
        return EXCEL_IMPORT_ERROR
    if step_type == SnowflakeImportStepPerformer.step_type():
        return SNOWFLAKE_IMPORT_ERROR
    return DATAFRAME_IMPORT_ERROR


def get_test_imports(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[int, str]:
    """
    Allows you to test the specific import steps, either by passing import steps
    or by passing an analysis name to check. 
    """
    # Get imports and flatten then, so we can iterate over them easily
    updated_step_import_data_list: Any = params['updated_step_import_data_list']
    all_imports = [_import for import_data in updated_step_import_data_list for _import in import_data['imports']]

    invalid_import_indexes: Dict[int, str] = dict()
    for index, _import in enumerate(all_imports):
        try:
            step = Step(_import['step_type'], 'fake_id', _import["params"])
            executed = step.set_prev_state_and_execute(steps_manager.steps_including_skipped[0].final_defined_state)
            if not executed:
                invalid_import_indexes[index] = get_import_error_for_step_type(_import['step_type'])
        except MitoError as e:
            invalid_import_indexes[index] = e.to_fix
        except:
            invalid_import_indexes[index] = get_import_error_for_step_type(_import['step_type'])
            

    return invalid_import_indexes



#!/usr/bin/env python

# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import DataframeFormat

class SetDataframeFormatStepPerformer(StepPerformer):
    """
    Allows you to set dataframe format.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'set_dataframe_format'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        df_format: DataframeFormat = get_param(params, 'df_format')
        
        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()

        post_state.df_formats[sheet_index] = df_format

        pandas_processing_time = perf_counter() - pandas_start_time


        return post_state, {
            'pandas_processing_time': pandas_processing_time,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            EmptyCodeChunk(
                prev_state, 
                post_state, 
                'Set dataframe format',
                'Set a dataframe format'
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')} # TODO: add the modified indexes here!


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
The Mito package, which contains functions for creating a Mito sheet. 

To generate a new sheet, simply run:

import mitosheet
mitosheet.sheet()

If running mitosheet.sheet() just prints text that looks like `MitoWidget(...`, then you need to 
install the JupyterLab extension manager by running:

jupyter labextension install @jupyter-widgets/jupyterlab-manager@2;

Run this command in the terminal where you installed Mito. It should take 5-10 minutes to complete.

Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

NOTE: if you have any issues with installation, please email jake@sagacollab.com
"""

import os
import pandas as pd

# Public interface we want users to rely on
from mitosheet.mito_backend import sheet
from mitosheet._version import __version__

# NOTE: We always export v1 sheet functions and types as unqualified exports, as we did
# this when we started Mito. This allows us to not break existing user analyses. We should
# never add to these -- see mitosheet/public/README.md for more info
from mitosheet.public.v1 import register_analysis
from mitosheet.public.v1.sheet_functions import *
from mitosheet.public.v1.sheet_functions.types import *
from mitosheet.public.v1.utils import flatten_column_header

# We export depricated utilities, so that users can still use them if they used to
from mitosheet.api.get_column_summary_graph import (
    filter_df_to_safe_size_external as filter_df_to_safe_size,
)
from mitosheet.step_performers.bulk_old_rename.deprecated_utils import (
    make_valid_header_external as make_valid_header,
)

# Make sure the user is initalized
from mitosheet.user import initialize_user
initialize_user()

# This function is only necessary for mitosheet3, as it is used
# in jlab3 to find the extension. It is not used in jlab2
def _jupyter_labextension_paths():
    """Called by jupyterlab to load the extension"""
    return [{"src": "labextension", "dest": "mitosheet"}]


def _jupyter_nbextension_paths():
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension and
    to install the widget

    Returns
    =======
    section: The section of the Jupyter Notebook Server to change.
        Must be 'notebook' for widget extensions
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Notebook copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Notebook copies
        from `src` directory into <jupyter path>/nbextensions/<dest> directory
        during widget installation
    require: Path to importable AMD Javascript module inside the
        <jupyter path>/nbextensions/<dest> directory
    """
    return [{
        'section': 'notebook',
        'src': 'nbextension',
        'dest': 'mitosheet',
        'require': 'mitosheet/extension'
    }]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests for errors during executing being rolled back properly
"""

import pytest

from mitosheet.tests.test_utils import create_mito_wrapper_with_data

INVALID_FORMULAS = [
    ("=1()"),
    ("=RIGHT1(100)"),
    ("=)"),
    ("=(NORK)"),
    ("=!!()"),
    ("=RIGHT(RIGHT(RIGHT()))"),
    ("=A + B & C - RIGHT(100)"),
    ("=hahahAHAH This is Random stuff")
]
@pytest.mark.parametrize("invalid_formula", INVALID_FORMULAS)
def test_invalid_formulas_dont_change_df(invalid_formula):
    """
    Makes sure that a column does not change when it is set to an
    invalid formula.
    """
    mito = create_mito_wrapper_with_data([123])
    mito.add_column(0, 'B')
    # Set the invalid formula
    mito.set_formula(invalid_formula, 0, 'B')
    # Shouldn't update the column at all!
    assert mito.get_value(0, 'B', 1) == 0
    





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
import traceback
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.user_defined_import_code_chunk import \
    UserDefinedImportCodeChunk
from mitosheet.errors import MitoError
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
import inspect

from mitosheet.types import UserDefinedImporterParamType


def get_user_defined_importer_param_type(f: Callable, param_name: str) -> UserDefinedImporterParamType:

    parameters = inspect.signature(f).parameters
    param_type = parameters[param_name].annotation

    if param_type == str:
        return 'str'
    elif param_type == int:
        return 'int'
    elif param_type == float:
        return 'float'
    elif param_type == bool:
        return 'bool'
    else:
        return 'any'


def get_param_names_to_types_for_importer(f: Callable) -> Dict[str, UserDefinedImporterParamType]:
    param_names_to_types = {}

    for name in inspect.signature(f).parameters:
        param_names_to_types[name] = get_user_defined_importer_param_type(f, name)

    return param_names_to_types

def get_user_defined_importers_for_frontend(state: Optional[State]) -> List[Any]:
    if state is None:
        return []

    return [
        {
            'name': f.__name__,
            'docstring': f.__doc__,
            'parameters': get_param_names_to_types_for_importer(f),
        }
        for f in state.user_defined_importers
    ]

def get_importer_params_and_type_and_value(f: Callable, frontend_params: Dict[str, str]) -> Dict[str, Tuple[UserDefinedImporterParamType, str]]:
    importer_params_and_type_and_value = {}

    for param_name, param_value in frontend_params.items():
        param_type = get_user_defined_importer_param_type(f, param_name)
        importer_params_and_type_and_value[param_name] = (param_type, param_value)
    
    return importer_params_and_type_and_value


def get_user_defined_importer_params_from_frontend_params(f: Callable, frontend_params: Dict[str, str]) -> Dict[str, Any]:
    user_defined_importer_params: Dict[str, Any] = {}

    for param_name, (param_type, param_value) in get_importer_params_and_type_and_value(f, frontend_params).items():
        try:
            if param_type == 'str':
                user_defined_importer_params[param_name] = param_value
            elif param_type == 'int':
                user_defined_importer_params[param_name] = int(param_value)
            elif param_type == 'float':
                user_defined_importer_params[param_name] = float(param_value)
            elif param_type == 'bool':
                user_defined_importer_params[param_name] = 'true' in param_value.lower()
            else:
                try:
                    user_defined_importer_params[param_name] = eval(param_value)
                except:
                    # If we cannot eval the result, it's likely a string, so we just pass it through
                    user_defined_importer_params[param_name] = param_value
        except:
            raise MitoError(
                'user_defined_importer_error',
                f"Importer {f.__name__} raised an error.",
                f"Parameter {param_name} with value {param_value} cannot be cast to type {param_type}. Please insert an appropriate value.",
                error_modal=False
            )

    return user_defined_importer_params



class UserDefinedImportStepPerformer(StepPerformer):
    """
    Allows a user to use one of the user_defined_importers.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'user_defined_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        importer: str = get_param(params, 'importer')
        importer_params: Dict[str, str] = get_param(params, 'importer_params')

        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()
        
        try:
            importer_function = next(f for f in post_state.user_defined_importers if f.__name__ == importer)
        except:
            raise MitoError(
                'user_defined_importer_not_found',
                f"Importer {importer} not found.",
                f"User defined importer {importer} not found. Please check that it is defined in the `importers` list passed to mitosheet.sheet.",
                error_modal=True
            )
        
        user_defined_importer_params = get_user_defined_importer_params_from_frontend_params(importer_function, importer_params)
    
        try:
            result = importer_function(**user_defined_importer_params)
        except:
            traceback_final_line = traceback.format_exc().splitlines()[-1]

            raise MitoError(
                'user_defined_importer_error',
                f"Importer {importer} raised an error.",
                f"User defined importer {importer} raised an error: {traceback_final_line}",
                error_modal=False
            )

        if isinstance(result, pd.DataFrame):
            new_dfs = [result]
        elif isinstance(result, list):
            new_dfs = result
        else:
            raise Exception(f"User defined importer {importer} must return a pandas dataframe or a list of pandas dataframes.")

        for df in new_dfs:
            post_state.add_df_to_state(
                df,
                DATAFRAME_SOURCE_IMPORTED,
            )

        pandas_processing_time = perf_counter() - pandas_start_time


        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'user_defined_importer_params': user_defined_importer_params,
            'result': {
                'num_new_dfs': len(new_dfs),
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            UserDefinedImportCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'importer'),
                get_param(execution_data if execution_data is not None else dict(), 'user_defined_importer_params')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_list_to_transpiled_code
from mitosheet.types import ColumnID

class MeltCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, id_var_column_ids: List[ColumnID], value_var_column_ids: List[ColumnID], include_value_vars: bool):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.id_var_column_ids = id_var_column_ids
        self.value_var_column_ids = value_var_column_ids
        self.include_value_vars = include_value_vars

        self.df_name = self.post_state.df_names[self.sheet_index]
        self.new_df_name = self.post_state.df_names[-1]

    def get_display_name(self) -> str:
        return 'Melt'
    
    def get_description_comment(self) -> str:
        return f"Unpivoted {self.df_name} into {self.new_df_name}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        id_vars = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, self.id_var_column_ids)
        value_vars = self.prev_state.column_ids.get_column_headers_by_ids(self.sheet_index, self.value_var_column_ids)

        value_vars = list(filter(lambda value_var: value_var not in id_vars, value_vars))

        param_string = f'id_vars={column_header_list_to_transpiled_code(id_vars)}'
        if self.include_value_vars:
            param_string += f', value_vars={column_header_list_to_transpiled_code(value_vars)}'

        return [f'{self.new_df_name} = {self.df_name}.melt({param_string})'], []
    
    def get_created_sheet_indexes(self) -> List[int]:
        return [len(self.post_state.dfs) - 1]
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Exports an error box that can be used to pass the errors we need. 

Further exports helper functions for creating errors
of different flavors. We wrap these functions in the same
MitoError box to avoid creating too many new classes!

See more about why we use errors here: 
https://stackoverflow.com/questions/16138232/is-it-a-good-practice-to-use-try-except-else-in-python
"""
import re
import traceback
from os.path import basename, normpath
from typing import Any, Collection, Iterable, List, Optional, Set, Union

from mitosheet.types import ColumnHeader, ColumnID, OperatorType, StateType


class MitoError(Exception):
    """
    An error that occurs during the processing of an editing event.
    """
    def __init__(self, type_: str, header: str, to_fix: str, error_modal: bool=True):
        """
        Creates a creation error. 

        type_: a string that is the error type. 
        header: the header of the error message to be displayed.
        to_fix: instructions on how to handle or fix the error.
        """
        self.type_ = type_ # we have an _ to avoid overwriting the build in type
        self.header = header
        self.to_fix = to_fix
        self.error_modal = error_modal
        self.traceback = get_recent_traceback() # record the most recent error when the error is created


def raise_error_if_column_ids_do_not_exist(step_display_name: str, state: StateType, sheet_index: int, column_ids: Union[List[ColumnID], ColumnID], error_modal: bool=True) -> None:
    df_name = state.df_names[sheet_index]

    # Make sure it's a list, if the user just passes a column header
    if not isinstance(column_ids, list):
        column_ids = [column_ids]

    for column_id in column_ids:
        if column_id not in state.column_ids.column_id_to_column_header[sheet_index]:
            raise MitoError(
                'no_column_error',
                'No Column Exists',
                f'A {step_display_name} step failed. The column "{column_id}" does not exist in {df_name}.',
                error_modal=error_modal
            )


def make_no_sheet_error(sheet_indexes: Set[int]) -> MitoError:
    """
    Helper function for creating a no_sheet_error.

    Occurs when a user edits a formula with a reference to a sheet that does not exist.
    """
    to_fix = f'There is an issue behind the scenes with that operation, we\'ll get it fixed as soon as possible!'

    return MitoError(
        'no_sheet_error', 
        'No Sheet Exists',
        to_fix
    )

def make_incompatible_merge_headers_error(error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a incompatible_merge_headers_error

    Occurs when a user merges on two dataframes, where one has a multi-index header
    and one does not.
    """
    to_fix = f'Invalid merge, as a dataframe has multi-index headers. To fix, edit your pivot table and flatten the headers.'

    return MitoError(
        'incompatible_merge_key_error', 
        'Incompatible Merge Key Types',
        to_fix,
        error_modal=error_modal
    )

def make_incompatible_merge_key_error(merge_key_one: Optional[ColumnHeader]=None, merge_key_one_dtype: Optional[str]=None, merge_key_two: Optional[ColumnHeader]=None, merge_key_two_dtype: Optional[str]=None, error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a incompatible_merge_key_error

    Occurs when a user merges on two keys that are different types
    """
    from mitosheet.column_headers import get_column_header_display

    if merge_key_one is not None and merge_key_two is not None:
        return MitoError(
            'incompatible_merge_key_error', 
            'Incompatible Merge Key Types',
            f'{get_column_header_display(merge_key_one)} ({merge_key_one_dtype}) and {get_column_header_display(merge_key_two)} ({merge_key_two_dtype}) have different types. Either pick new keys or cast their types.',
            error_modal=error_modal
        )
    
    to_fix = f'You cannot merge using keys of different types. Either pick new keys or cast their types.'
    return MitoError(
        'incompatible_merge_key_error', 
        'Incompatible Merge Key Types',
        to_fix,
        error_modal=error_modal
    )

def make_no_column_error(column_headers: Collection[ColumnHeader], error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a no_column_error.

    Occurs when a user edits a formula with a reference to a column that does not exist.
    """
    if len(column_headers) == 1:
        to_fix = f'Sorry, there is no column with the name {next(iter(column_headers))}. Did you type it correctly?'
    else:
        to_fix = f'Sorry, there are no column with the names {", ".join(map(str, column_headers))}. Did you type them correctly?'

    return MitoError(
        'no_column_error', 
        'No Column Exists',
        to_fix,
        error_modal=error_modal
    )

def make_column_exists_error(column_header: ColumnHeader) -> MitoError:
    """
    Helper function for creating a column_exists_error.

    Occurs when:
    -  the user adds a column that already exists in the dataframe.
    """
    return MitoError(
        'column_exists_error', 
        'Column Already Exists',
        f'Sorry, a column already exists with the name {column_header}. Try picking a different name!'
    )

def make_columns_exists_error(column_headers: Iterable[ColumnHeader]) -> MitoError:
    """
    Helper function for creating a column_exists_error.

    Occurs when:
    -  the user adds a column(s) that already exists in the dataframe.
    """
    return MitoError(
        'columns_exists_error', 
        'Columns Already Exist',
        f'{(", ").join(map(str, column_headers))} alread exist. Try renaming these columns!'
    )

def make_invalid_formula_error(formula: str, to_fix: Optional[str]=None, error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a invalid_formula_error.

    Occurs when:
    -  the user edits a cell to an expression that cannot be parsed.
    """
    if to_fix is None:
        to_fix = f'Sorry, the formula \'{formula}\' is invalid. Did you type it correctly?'

    return MitoError(
        'invalid_formula_error',
        'Invalid Formula',
        to_fix, 
        error_modal=error_modal
    )

def make_invalid_formula_after_update_error(error_modal: bool=False) -> MitoError:
    """
    Helper function for creating a invalid_formula_after_update_error.

    Occurs when a user edits a user tries to resubmit a formula that used to be
    valid and is no longer valid - which must mean they changed something they 
    rely on.
    """
  
    return MitoError(
        'invalid_formula_after_update_error', 
        'Formula is Now Invalid',
        'Renaming or deleting other columns in this dataset has made this formula invalid. Please update this formula.',
        error_modal=error_modal
    )


def make_cast_value_to_type_error(value: str, column_dtype: str, error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a cast_value_to_type_error.

    Occurs when the user edits a cell to a specific value that cannot be cast to the 
    type of the column it is in.

    """
    return MitoError(
        'cast_value_to_type_error',
        'Invalid Value for ' + column_dtype + 'Series',
        'The value ' + value + ' could not be cast to the type ' + column_dtype + '. Please enter a different value.' , 
        error_modal=error_modal
    )

def make_circular_reference_error(error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a circular_reference_error.

    Occurs when:
    -  an edit introduces a circular reference.
    """
    return MitoError(
        'circular_reference_error',
        'Circular Reference',
        f'Sorry, circular references are not supported currently.',
        error_modal=error_modal
    )

def make_invalid_column_headers_error(column_headers: List[ColumnHeader]) -> MitoError:
    """
    Helper function for creating a invalid_column_headers_error.

    Occurs when:
    -  a user creates (or renames) a column(s) that has an invalid name.
    """
    to_fix = f'All headers in the dataframe must contain at least one letter and no symbols other than numbers and "_". Invalid headers: {", ".join(map(str, column_headers))}'

    return MitoError(
        'invalid_column_header_error',
        'Invalid Column Header',
        to_fix
    )

def make_function_error(function_name: str, error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a function_error.

    Occurs when there is an error inside a specific sheet function that
    is running
    """
    # TODO: this to_fix message is _just terrible_
    return MitoError(
        'function_error',
        'Error in ' + function_name,
        f'Sorry, an unknown error occured in the ' + function_name + ' function. See the documentation for help.',
        error_modal=error_modal
    )


def make_execution_error(error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a execution_error.

    Occurs when:
    -  running an existing analysis on the parameters passed to the mito sheet throw an error.
    """
    # TODO: this to_fix message is _just terrible_
    return MitoError(
        'execution_error',
        'Execution Error',
        f'Sorry, there was an error during executing this code.',
        error_modal=error_modal
    )

def make_function_execution_error(function: str) -> MitoError:
    """
    Helper function for creating a function_execution_error.

    Occurs when:
    -  An error occurs inside of a mito sheet function
    """
    # TODO: this to_fix message is _just terrible_
    return MitoError(
        'execution_error',
        f'Error Executing {function}',
        f'Sorry, there was an error in the {function}. Please check the documentation to make sure you called it correctly.'
    )

def make_unsupported_function_error(functions: Set[str], error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a unsupported_function_error.

    Occurs when:
    -  the user's new formula containing a function that is not supported by mito.
    """
    # TODO: we could fix up this to_fix, lol.
    if len(functions) == 1:
        to_fix = f'Sorry, mito does not currently support the function {next(iter(functions))}. To request it, shoot us an email!'
    else:
        to_fix = f'Sorry, mito does not currently support the functions {", ".join(functions)}. To request it, shoot us an email!'

    return MitoError(
        'unsupported_function_error',
        'Unsupported Function',
        to_fix,
        error_modal=error_modal
    )


def make_invalid_column_delete_error(column_headers: Collection[ColumnHeader], dependents: Optional[Collection[ColumnHeader]]=None) -> MitoError:
    """
    Helper function for creating a invalid_column_delete_error.

    Occurs when:
    - the user deletes a column that is referenced by other columns
    - the user deletes a column that does not exist
    """
    if dependents is not None:
        header = 'Column Has Dependents'
        to_fix = f'{(", ").join(map(str, column_headers))} cannot be deleted, as {"they are" if len(column_headers) > 1 else "it is"} referenced in {(", ".join(map(str, dependents)))}. Please remove these references before deleting.'
    else:
        header = 'Column Does not exist'
        to_fix = f'{(", ").join(map(str, column_headers))} cannot be deleted, as not all of these columns exist.'
   

    return MitoError(
        'invalid_column_delete_error',
        header,
        to_fix
    )

def make_invalid_arguments_error(function: str) -> MitoError:
    """
    Helper function for creating a invalid_arguments_error.

    Occurs when:
    -  the user calls a sheet function with incorrect arguments
    """
    return MitoError(
        'invalid_arguments_error',
        f'Invalid Arguments to {function}',
        f'It looks like you passed the wrong arguments to {function}. Try checking out the documentation to see correct usage!'
    )

def make_invalid_filter_error(filter_value: Any, correct_type: str) -> MitoError:
    """
    Helper function for creating a invalid_filter_error.

    Occurs when:
    - A user tries to create a filter with an invalid value (e.g. a string, when filtering numbers
      or a non-date when filtering dates).
    """

    if correct_type == 'datetime':
        correct_format = 'as YYYY-MM-DD.'
    elif correct_type == 'number':
        correct_format = 'that is a number.'
    else:
        correct_format = 'that is valid.'

    return MitoError(
        'invalid_filter_error',
        f'Invalid Filter',
        f'Sorry, the value {filter_value} is not a valid value for that {correct_type} filter. Please enter a value {correct_format}!'
    )

def make_duplicated_column_headers_error(duplicated_headers: List[ColumnHeader]) -> ValueError:
    """
    Helper function for creating a invalid_filter_error.

    Occurs when:
    - A user tries to pass a column with duplicated column headers
    """

    return ValueError(
        f'The column headers {duplicated_headers} are duplicated in the dataframe. Please only pass dataframes with unique column names.'
    )

def make_invalid_sort_error(column_header: ColumnHeader) -> MitoError:
    """
    Helper function for creating a invalid_sort_error.

    Occurs when:
    - A user tries to sort a dataframe on a column of mixed types 
    (e.g. a column with strings and floats).
    """

    return MitoError(
        'invalid_sort_error',
        f'Invalid Sort',
        f'Sorry, the column {column_header} has mixed data types. Please make sure the column has one datatype before trying to sort.'
    )

def make_df_exists_error(df_name: str) -> MitoError:
    """
    Helper function for creating a df_exists_error.

    Occurs when:
    - A user tries to create a dataframe with the name of an already existing
    dataframe in the sheet.
    """

    return MitoError(
        'df_exists_error',
        f'Dataframe Exists',
        f'Sorry, the dataframe {df_name} already exists. Please pick a different name.'
    )

def make_invalid_column_type_change_error(column_header: ColumnHeader, old_dtype: str, new_dtype: str) -> MitoError:
    """
    Helper function for creating a invalid_column_type_change_error.

    Occurs when:
    - A user tries to change the type of one column to an incompatible type (e.g. from
      a number to a datetime, which is impossible).
    """

    return MitoError(
        'invalid_column_type_change_error',
        f'Invalid Type Change',
        f'Sorry, the column {column_header} has a type {old_dtype}, which cannot be changed to the type {new_dtype}.'
    )

def make_invalid_pivot_error() -> MitoError:
    """
    Helper function for creating a invalid_pivot_error.

    Occurs when:
    -  the user runs a pivot that is invalid in some way.
    """
    # TODO: this to_fix message is _just terrible_
    return MitoError(
        'invalid_pivot_error',
        'Pivot Error',
        f'Sorry, there was an error computing your pivot. Please try a different pivot!',
        error_modal=False
    )

def make_invalid_pivot_filter_error(column_header: ColumnHeader, filter_condition: str) -> MitoError:
    """
    Helper function for creating a invalid_pivot_filter_error.

    Occurs when:
    -  the user runs a pivot that is invalid in some way.
    """
    # TODO: this to_fix message is _just terrible_
    return MitoError(
        'invalid_pivot_filter_error',
        'Pivot Error',
        f'Sorry, the filter {filter_condition} on {column_header} is invalid. This is likely due to incompatible dtypes. Try changing the filter condition or column dtype.',
        error_modal=False
    )


def make_is_directory_error(file_name: str) -> MitoError:
    """
    Helper function for creating directory_error

    Occurs when:
    - the user tries to import a directory in a simple import
    """
    return MitoError(
        'directory_error',
        f'Change Path to a File',
        f'Sorry, the path {file_name} is a directory. Please pass in a path to a file to import it into Mito.'
    )


def make_no_analysis_error(analysis_id: str, error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a no_analysis_error.

    Occurs when a user tries to replay an analysis that they don't have access to in 
    their .mito folder
    """
    to_fix = f'When you call mitosheet.sheet() and there is a code cell below that contains an analysis ID, \
    Mito tries to replay that analysis. However, you do not have access to the analysis {analysis_id} so Mito is unable to replay it. \
    Add a new code cell below the mitosheet.sheet() call to start a new analysis.' 
    
    return MitoError(
        'no_analysis_error', 
        "You don't have access to the analysis",
        to_fix,
        error_modal=error_modal
    )

def make_invalid_update_imports_error() -> MitoError:
    """
    Helper function for creating a invalid_update_imports_error.

    Occurs when the user tries to update update the imports in an analysis in a way
    that leads to no importing errors, but other errors as the structure of the data
    has changed from the original dataset.
    """
    
    return MitoError(
        'invalid_update_imports_error', 
        "Cannot update imports",
        'The newly imported data does not have the same structure as the original data. Likely, a column used in the analysis is not part of the new data set.',
        error_modal=False
    )

def make_invalid_promote_row_to_header(error_modal: bool=True) -> MitoError:
    """
    Helper function for creating a invalid_promote_row_to_header.

    Occurs when a user tries to promote a row to header that has duplicated
    values within it.
    """
    to_fix = f'This row has duplicated values in it. As such, making this row a header would make column references ambigious, so we cannot make it a header row.' 
    
    return MitoError(
        'no_analysis_error', 
        "Cannot create duplicate headers",
        to_fix,
        error_modal=error_modal
    )

def make_invalid_simple_import_error(error_modal: bool=False) -> MitoError:
    """
    Helper function for creating an invalid_simple_import_error.

    Occurs when a user tries to simple import and it fails
    """
    to_fix = f'We were unable to automatically determine the import configurations, like delimiter and encoding. Update the import configuration.' 
    
    return MitoError(
        'invalid_simple_import_error', 
        "Cannot Determine File Data",
        to_fix,
        error_modal=error_modal
    )

def make_invalid_snowflake_import_error(exception: Exception) -> MitoError:
    """
    Helper function for creating an invalid_snowflake_import_error.

    Occures when the user tries to snowflake import and it fails.
    """
    return MitoError(
        'invalid_snowflake_import_error',
        'Cannot Query Database',
        f'{exception}',
        False
    )

def make_invalid_snowflake_credentials_error() -> MitoError:
    """
    Helper function for creating an invalid_snowflake_credentials_error.

    Occures when the user tries to snowflake imports, but their credentials are invalid.
    """
    return MitoError(
        'invalid_snowflake_credentials_error',
        'Invalid Snowflake Credentials',
        'Your Snowflake connection credentials are invalid. Please enter valid Snowflake credentials and try again.',
        False
    )

def make_file_not_found_error(file_name: str) -> MitoError:
    """
    Helper function for creating a file_not_found_error.

    Occurs when a user tries to import a file that does not exist
    """
    
    return MitoError(
        'file_not_found_error', 
        "File does not exist",
        f'{basename(normpath(file_name))} does not exist.',
        error_modal=False
    )

def make_dataframe_not_found_error(df_name: str) -> MitoError:
    """
    Occurs when a user tries to import a dataframe that does not exist
    """
    
    return MitoError(
        'dataframe_not_found_error', 
        "File does not exist",
        f'{df_name} is not defined.'
    )

ARG_FULL_NAME = {
    'int': 'number',
    'float': 'number',
    'str': 'string',
    'bool': 'boolean'
}

def make_operator_type_error(operator: str, arg_one_type: str, arg_two_type: str) -> MitoError:
    """
    Helper function for creating a operator_type_error.

    Occurs when:
    - user uses an operator to add/mul/div (etc) invalid types. E.g. they do: 1 + "hi",
      or pd.Series(['hi']) / 10.
    """

    # Provide a helpful error message, if possible, for the case of a number and a string
    # (which is the most common of these errors).
    if (arg_one_type == 'str' and (arg_two_type == 'int' or arg_two_type == 'float')) or \
        (arg_two_type == 'str' and (arg_one_type == 'int' or arg_one_type == 'float')):
        to_fix = f'Sorry, you tried to {operator} on a string and a number. Please wrap the number argument in a call to the VALUE function.'
    else:
        # Otherwise, handle errors in a non-specific way - this is a catch all for the rest of operator type errors
        type_one_full_name = ARG_FULL_NAME[arg_one_type] if arg_one_type in ARG_FULL_NAME else arg_one_type
        type_two_full_name = ARG_FULL_NAME[arg_two_type] if arg_two_type in ARG_FULL_NAME else arg_two_type
        # TODO: once we have a step that changes the column type, we should update this error to be more helpful
        to_fix = f'Sorry, you tried to use {operator} on a {type_one_full_name} and a {type_two_full_name}. Please change the type of one of these columns to be compatible.'

    return MitoError(
        'operator_type_error',
        f'Error with {operator}',
        to_fix
    )


def make_invalid_range_error(range: str, error_modal: bool) -> MitoError:
    """
    Helper function for creating a invalid_range_error.

    Occurs when the user does not 
    """
    return MitoError(
        'invalid_range_error',
        'Invalid Range',
        f'Range {range} is not in the format A1:B3. Please update to COLUMNROW:COLUMNROW.',
        error_modal=error_modal
    )

def make_exec_error(e: Exception) -> MitoError:
    """
    Helper function for creating a make_exec_error, which captures an error that
    occurs inside of exec (from AI generated code).
    """

    return MitoError(
        'make_exec_error',
        'Exec Error',
        f'{type(e).__name__}: {e}',
        error_modal=False
    )


def make_range_not_found_error(value: Union[str, int, float, bool], error_modal: bool) -> MitoError:
    """
    Helper function for creating a range_not_found_error.

    Occurs when the user does not give a value that is found in the sheet.
    """
    return MitoError(
        'range_not_found_error',
        'Value Not Found',
        f'Value {value} was not found in the specified sheet. Please ensure the passed value is in this sheet.',
        error_modal=error_modal
    )

def get_recent_traceback() -> str:
    """
    Helper function that returns the most recent traceback, with the file paths
    stripped to remove all but the mitosheet file names for ease in debugging.
    
    Inspired by https://stackoverflow.com/questions/25272368/hide-file-from-traceback
    """
    return re.sub(r'File ".*[\\/]([^\\/]+.py)"', r'File "\1"', traceback.format_exc())

def get_recent_traceback_as_list() -> List[str]:
    """
    Helper function for getting the traceback of the most recent error.
    
    This, in turns, gives us a stack trace for any error that we want
    to log with detail, in a format that is sendable to mixpanel.

    We get this as a list of strings, as mixpanel truncates strings at 255 chars
    and thus we avoid things getting chopped this way.
    """
    # NOTE: We ignore empty lines, as they add no information
    return [line for line in get_recent_traceback().split('\n') if line != '']






from typing import Callable, List, Optional, Tuple, TypeVar, Union
import numpy as np
from datetime import datetime, timedelta

import pandas as pd

from mitosheet.public.v3.rolling_range import RollingRange


PrimitiveType = TypeVar('PrimitiveType', bound=Union[str, float, int, bool, datetime, timedelta])

ResultType = Union[pd.Series, PrimitiveType]


def __get_default_value_if_value_is_none_or_nan(value: PrimitiveType, default_value: PrimitiveType) -> PrimitiveType:
    if value is None or (isinstance(value, float) and np.isnan(value)):
        return default_value
    
    return value


def __get_new_result_series_or_primitive_helper(
        default_value: PrimitiveType,
        previous_result: ResultType, 
        new_value: ResultType,
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This helper function is just responsible for combining two previous results into one new
    result. 

    It does so by:
    1. Filling NaN values with the default value.
    2. Delegating to the get_new_result_from_primitive_values and get_new_result_from_series functions
    """
    if isinstance(previous_result, pd.Series):
        if isinstance(new_value, pd.Series):
            result =  get_new_result_from_series(previous_result.fillna(default_value), new_value.fillna(default_value))
            return result
        else:
            return previous_result.apply(lambda v: get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(v, default_value), 
                __get_default_value_if_value_is_none_or_nan(new_value, default_value)
            ))
    else:
        if isinstance(new_value, pd.Series):
            return new_value.apply(lambda v: get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(previous_result, default_value), 
                __get_default_value_if_value_is_none_or_nan(v, default_value)
            ))
        else:
            return get_new_result_from_primitive_values(
                __get_default_value_if_value_is_none_or_nan(previous_result, default_value), 
                __get_default_value_if_value_is_none_or_nan(new_value, default_value), 
            )


def __get_new_result_series_or_primitive(
        default_value: PrimitiveType,
        previous_result: ResultType,
        arg: Union[PrimitiveType, None, pd.Series, RollingRange, pd.DataFrame], 
        get_primitive_value_from_dataframe: Callable[[pd.DataFrame], PrimitiveType],
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This helper function does the preprocessing for a single arg, and then combines it
    with the previous result to get the new result.
    """

    def get_new_result(_previous_result: ResultType, new_value: ResultType) -> ResultType:
        return __get_new_result_series_or_primitive_helper(
            default_value,
            _previous_result,
            new_value,
            get_new_result_from_primitive_values,
            get_new_result_from_series
        )

    if isinstance(arg, pd.DataFrame):
        reduced_df = get_primitive_value_from_dataframe(arg)
        return get_new_result(previous_result, reduced_df)

    elif isinstance(arg, RollingRange):
        new_series = arg.apply(lambda df: get_primitive_value_from_dataframe(df))
        return get_new_result(previous_result, new_series)
        
    elif isinstance(arg, pd.Series):
        return get_new_result(previous_result, arg)
        
    elif arg is not None:
        return get_new_result(previous_result, arg)
    
    # Otherwise the arg is None, so we just skip it
    return previous_result


def get_final_result_series_or_primitive(
        default_value: PrimitiveType,
        argv: Tuple[Union[PrimitiveType, None, pd.Series, RollingRange, pd.DataFrame], ...], 
        get_primitive_value_from_dataframe: Callable[[pd.DataFrame], PrimitiveType],
        get_new_result_from_primitive_values: Callable[[PrimitiveType, PrimitiveType], PrimitiveType],
        get_new_result_from_series: Callable[[pd.Series, pd.Series], pd.Series]
    ) -> ResultType:
    """
    This function is the main workhorse of many sheet functions that fit a common pattern:

    1. The start with a default result (default_value).
    2. They update the result with each arg in two steps, preprocessing the arg and then combining that with the result
    3. Preprocessing the arg:
        - For dataframe values, they turned into primitive values with get_primitive_value_from_dataframe
        - For rolling ranges, they are turned into series using repeated application of get_primitive_value_from_dataframe
    4. Combining with the previous result. We are either combining two primtiive values, a primitive value and a series, or two series
        - If combining two primitive values, we combine with get_new_result_from_primitive_values
        - If combining a primitive value and a series, we use a .apply on the series with get_new_result_from_primitive_values
        - If combining two series, then we use get_new_result_from_series

    This makes it much easier to specify many functions like AND, SUM, OR, etc. 

    Notably, to make get_new_result_from_primitive_values and get_new_result_from_series easier, we fill all NaN and None
    values in the primitive values and series results being combined with the default_value before sending them to 
    these functions in step (4).
    """

    result: ResultType = default_value

    for arg in argv:
        result = __get_new_result_series_or_primitive(
            default_value,
            result,
            arg,
            get_primitive_value_from_dataframe,
            get_new_result_from_primitive_values,
            get_new_result_from_series
        )

    return result 
    

def get_index_from_series(*args: Union[PrimitiveType, pd.Series]) -> pd.Index:
    if any(isinstance(arg, pd.Series) for arg in args):
        # Search for the first series and use its index
        return next(arg.index for arg in args if isinstance(arg, pd.Series))
    else:
        raise Exception("No series found in args")
    

def get_series_from_primitive_or_series(
        arg: Union[PrimitiveType, pd.Series], 
        index: pd.Index
    ) -> pd.Series:
    if isinstance(arg, pd.Series):
        return arg
    else:
        return pd.Series([arg] * len(index), index=index)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.types import ColumnID

class TransposeCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index

        self.df_name = self.post_state.df_names[self.sheet_index]
        self.transposed_df_name = self.post_state.df_names[-1]

    def get_display_name(self) -> str:
        return 'Transpose'
    
    def get_description_comment(self) -> str:
        return f"Transposed {self.df_name} into {self.transposed_df_name}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        return [f'{self.transposed_df_name} = {self.df_name}.T'], []

    def get_created_sheet_indexes(self) -> List[int]:
        return [len(self.post_state.dfs) - 1]
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict, List, Tuple
from mitosheet.code_chunks.code_chunk_utils import get_code_chunks
from mitosheet.types import StepsManagerType
from mitosheet.updates.args_update import is_string_arg_to_mitosheet_call


def get_parameterizable_params(params: Dict[str, Any], steps_manager: StepsManagerType) -> List[Tuple[str, str, str]]:


        all_parameterizable_params: List[Tuple[str, str, str]] = []

        # First, get the original arguments to the mitosheet - we only let you parameterize df names for now
        for arg in steps_manager.original_args_raw_strings:
                if not is_string_arg_to_mitosheet_call(arg):
                        all_parameterizable_params.append((arg, 'df_name', "Dataframe"))
    
        # Get optimized code chunk, and get their parameterizable params
        code_chunks = get_code_chunks(steps_manager.steps_including_skipped[:steps_manager.curr_step_idx + 1], optimize=True)

        for code_chunk in code_chunks:
                parameterizable_params = code_chunk.get_parameterizable_params()
                all_parameterizable_params.extend(parameterizable_params)

        return all_parameterizable_params



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Schemas for the user.json file. As we evolve the format overtime,
we should include them here, so that we can track what upgrades
are required / necessary.

NOTE: USER_JSON_DEFAULT is the default user.json object for the currrent
version always.
"""

from datetime import datetime
from mitosheet.experiments.experiment_utils import get_new_experiment
from mitosheet.utils import get_random_id
from mitosheet._version import __version__

# Some helpful constants
GITHUB_ACTION_ID = 'github_action'
GITHUB_ACTION_EMAIL = 'github@action.com'

# Old field names
UJ_INTENDED_BEHAVIOR = 'intended_behavior'
UJ_CLOSED_FEEDBACK = 'closed_feedback'
UJ_MITOSHEET_LAST_FIVE_USAGES = 'mitosheet_last_five_usages'

# Current field names
UJ_USER_JSON_VERSION = 'user_json_version'
UJ_STATIC_USER_ID = 'static_user_id'
UJ_USER_SALT = 'user_salt'
UJ_USER_EMAIL = 'user_email'
UJ_RECEIVED_TOURS = 'received_tours'
UJ_FEEDBACKS = 'feedbacks'
UJ_FEEDBACKS_V2 = 'feedbacks_v2'
UJ_MITOSHEET_CURRENT_VERSION = 'mitosheet_current_version'
UJ_MITOSHEET_LAST_UPGRADED_DATE = 'mitosheet_last_upgraded_date'
UJ_MITOSHEET_LAST_FIFTY_USAGES = 'mitosheet_last_fifty_usages'
UJ_MITOSHEET_TELEMETRY = 'mitosheet_telemetry'
UJ_MITOSHEET_PRO = 'mitosheet_pro'
UJ_MITOSHEET_ENTERPRISE = 'mitosheet_enterprise'
UJ_EXPERIMENT = 'experiment'
UJ_RECEIVED_CHECKLISTS = 'received_checklists'
UJ_AI_PRIVACY_POLICY = 'ai_privacy_policy'
UJ_AI_MITO_API_NUM_USAGES = 'ai_mito_api_num_usages'


# This was the final version of the user.json with user_json_version == 1
# and since we added fields to it over time, we make sure that all users 
# upgrading from version 1 actually get all the fields they need before 
# being upgraded
# NOTE: tests ensure we cannot change this object
USER_JSON_VERSION_1 = {
    UJ_USER_JSON_VERSION: 1,
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    UJ_USER_EMAIL: '',
    # A list of actions the user intends to do on the tool, which they fill
    # out when they sign up
    UJ_INTENDED_BEHAVIOR: [],
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    # If the user opted out of feedback, we store that they opted out, so
    # that we don't bombard them with feedback
    UJ_CLOSED_FEEDBACK: False,
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIVE_USAGES: [datetime.today().strftime('%Y-%m-%d')]
}

USER_JSON_VERSION_2 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 2,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')]
}

USER_JSON_VERSION_3 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 3,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True
}

"""
The only difference between version 3 and 4 is the addition of the UJ_FEEDBACKS_V2. 
We moved to this format because we wanted to insist on more structure to our feedback, 
allowing us to more intelligently extract information from the feedback object if we need to. 

It has the format:
{
	"pivot_edit_usage_triggered": [
		{
			"date": "2021-11-16",
			"num_usages": 1
			"questions_and_answers": [
				{
					"question": "how successful was that pivot table?",
					"answer":  5
				},
				{
					"question": "what was your experience like",
					"answer":  "It was difficult because ...", 
				}
			]
		},
		{
			"date": "2021-11-17",
			"num_usages": 2
			"questions_and_answers": [
				{
					"question": "how successful was that pivot table?",
					"answer":  5
				},
				{
					"question": "what was your experience like",
					"answer":  "It was difficult because ...", 
				}
			]
		}
	],
	"other_feedback_id": [
		...
	]
}
"""
USER_JSON_VERSION_4 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 4,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True
}

"""
Adds UJ_MITOSHEET_PRO, and sets to equal to False. The upgrade step
sets it to False only if the field is not already there.
"""
USER_JSON_VERSION_5 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 5,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False
}

"""
Adds UJ_EXPERIMENTS, and sets it equal to an empty list. This is a list of experiments
that the user has been assigned to. It has the format [{'experiment_id': '<ID>', 'variant': '<VARIANT>'}]
"""
USER_JSON_VERSION_6 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 6,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_EXPERIMENT: get_new_experiment()
}

"""
Adds UJ_RECEIVED_CHECKLISTS, which is a list of checklists that the user has received,
and the steps for each checklist they have been through, with the format:
{"onboarding_checklist": ['signup', 'import', 'filter', 'pivot', 'graph']}
"""
USER_JSON_VERSION_7 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 7,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_EXPERIMENT: get_new_experiment(),
    UJ_RECEIVED_CHECKLISTS: {}
}

"""
Adds UJ_MITOSHEET_ENTERPRISE and sets it equal to False
"""
USER_JSON_VERSION_8 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 8,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_MITOSHEET_ENTERPRISE: False,
    UJ_EXPERIMENT: get_new_experiment(),
    UJ_RECEIVED_CHECKLISTS: {}
}

"""
Adds UJ_AI_PRIVACY_POLICY and sets it equal to False
"""
USER_JSON_VERSION_9 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 9,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_MITOSHEET_ENTERPRISE: False,
    UJ_EXPERIMENT: get_new_experiment(),
    UJ_RECEIVED_CHECKLISTS: {},
    UJ_AI_PRIVACY_POLICY: False
}

"""
Adds UJ_AI_MITO_API_NUM_USAGES and sets it equal to 0
"""
USER_JSON_VERSION_10 = {
    # The new version of the user json object
    UJ_USER_JSON_VERSION: 10,
    # The static id of the user
    UJ_STATIC_USER_ID: get_random_id(),
    # A random secret that the user can use as salt when hashing things
    UJ_USER_SALT: get_random_id(),
    # Email of the user
    UJ_USER_EMAIL: '',
    # Tours that the user has received
    UJ_RECEIVED_TOURS: [],
    # A list of all the feedback the user has given
    UJ_FEEDBACKS: [],
    UJ_FEEDBACKS_V2: {},
    UJ_MITOSHEET_CURRENT_VERSION: __version__,
    UJ_MITOSHEET_LAST_UPGRADED_DATE: datetime.today().strftime('%Y-%m-%d'),
    UJ_MITOSHEET_LAST_FIFTY_USAGES: [datetime.today().strftime('%Y-%m-%d')],
    UJ_MITOSHEET_TELEMETRY: True,
    UJ_MITOSHEET_PRO: False,
    UJ_MITOSHEET_ENTERPRISE: False,
    UJ_EXPERIMENT: get_new_experiment(),
    UJ_RECEIVED_CHECKLISTS: {},
    UJ_AI_PRIVACY_POLICY: False,
    UJ_AI_MITO_API_NUM_USAGES: 0
}


# This is the most up to date user json, and you must update it when
# you add a new schema
USER_JSON_DEFAULT = USER_JSON_VERSION_10



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_list_to_transpiled_code, column_header_to_transpiled_code
from mitosheet.types import ColumnHeader, ColumnID

class OneHotEncodingCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, new_column_headers: List[ColumnHeader]):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_id = column_id
        self.new_column_headers = new_column_headers

    def get_display_name(self) -> str:
        return 'One Hot Encoding'
    
    def get_description_comment(self) -> str:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        return f'One-hot Encoded {column_header}'
        
    def get_code(self) -> Tuple[List[str], List[str]]:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        df_name = self.post_state.df_names[self.sheet_index]
        column_index = self.post_state.dfs[self.sheet_index].columns.tolist().index(column_header)

        transpiled_column_header = column_header_to_transpiled_code(column_header)
        new_transpiled_column_headers = column_header_list_to_transpiled_code(self.new_column_headers)

        encode_line = f'{df_name}[{new_transpiled_column_headers}] = pd.get_dummies({df_name}[{transpiled_column_header}])'
        reorder_columns_line = f'{df_name} = {df_name}[{df_name}.columns[:{column_index + 1}].tolist() + {new_transpiled_column_headers} + {df_name}.columns[{column_index + 1}:-{len(self.new_column_headers)}].tolist()]'

        return [
            encode_line,
            reorder_columns_line   
        ], ['import pandas as pd']

    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import SimpleImportCodeChunk
from mitosheet.code_chunks.step_performers.import_steps.excel_import_code_chunk import ExcelImportCodeChunk
from mitosheet.code_chunks.step_performers.import_steps.excel_range_import_code_chunk import ExcelRangeImportCodeChunk
from mitosheet.code_chunks.snowflake_import_code_chunk import SnowflakeImportCodeChunk
from mitosheet.code_chunks.step_performers.pivot_code_chunk import PivotCodeChunk
from mitosheet.code_chunks.step_performers.merge_code_chunk import MergeCodeChunk
from mitosheet.code_chunks.step_performers.dataframe_steps.dataframe_duplicate_code_chunk import DataframeDuplicateCodeChunk
from mitosheet.state import State


class DataframeRenameCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, old_dataframe_name: str, new_dataframe_name: str):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.old_dataframe_name = old_dataframe_name
        self.new_dataframe_name = new_dataframe_name

    def get_display_name(self) -> str:
        return 'Renamed Dataframe'
    
    def get_description_comment(self) -> str:
        return f'Renamed {self.old_dataframe_name} to {self.new_dataframe_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        if self.old_dataframe_name == self.new_dataframe_name:
            return [], []

        return [f'{self.post_state.df_names[self.sheet_index]} = {self.old_dataframe_name}'], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_left_with_dataframe_rename(self, dataframe_rename_code_chunk: "DataframeRenameCodeChunk") -> Optional["DataframeRenameCodeChunk"]:
        if self.sheet_index != dataframe_rename_code_chunk.sheet_index:
            return None
        
        return DataframeRenameCodeChunk(
            dataframe_rename_code_chunk.prev_state,
            self.post_state,
            dataframe_rename_code_chunk.sheet_index,
            dataframe_rename_code_chunk.old_dataframe_name,
            self.new_dataframe_name
        )

    def _combine_left_with_pivot_code_chunk(self, pivot_code_chunk: PivotCodeChunk) -> Optional[CodeChunk]:
        destination_sheet_index = pivot_code_chunk.destination_sheet_index
        if destination_sheet_index is None:
            destination_sheet_index = len(pivot_code_chunk.post_state.dfs) - 1

        if destination_sheet_index != self.sheet_index:
            return None

        return PivotCodeChunk(
            pivot_code_chunk.prev_state,
            self.post_state,
            pivot_code_chunk.sheet_index,
            pivot_code_chunk.destination_sheet_index,
            pivot_code_chunk.pivot_rows_column_ids_with_transforms,
            pivot_code_chunk.pivot_columns_column_ids_with_transforms,
            pivot_code_chunk.pivot_filters_ids,
            pivot_code_chunk.values_column_ids_map,
            pivot_code_chunk.flatten_column_headers,
            pivot_code_chunk.was_series,
            pivot_code_chunk.public_interface_version
        )

    def _combine_left_with_merge_code_chunk(self, merge_code_chunk: MergeCodeChunk) -> Optional[CodeChunk]:
        destination_sheet_index = len(merge_code_chunk.post_state.dfs) - 1
        if destination_sheet_index != self.sheet_index:
            return None

        return MergeCodeChunk(
            merge_code_chunk.prev_state,
            self.post_state,
            merge_code_chunk.how,
            merge_code_chunk.sheet_index_one,
            merge_code_chunk.sheet_index_two,
            merge_code_chunk.merge_key_column_ids,
            merge_code_chunk.selected_column_ids_one,
            merge_code_chunk.selected_column_ids_two,
        )

    def _combine_left_with_import_code_chunk(
            self, 
            import_code_chunk: Union[SimpleImportCodeChunk, ExcelImportCodeChunk, ExcelRangeImportCodeChunk, SnowflakeImportCodeChunk]
        ) -> Optional[CodeChunk]:
        created_sheet_indexes = import_code_chunk.get_created_sheet_indexes()
        if created_sheet_indexes is None or self.sheet_index not in created_sheet_indexes:
            return None

        new_import_chunk = deepcopy(import_code_chunk)
        new_import_chunk.post_state = self.post_state
        return new_import_chunk

    def _combine_left_with_dataframe_duplicate(
        self, 
        dataframe_duplicate_code_chunk: DataframeDuplicateCodeChunk
    ) -> Optional[CodeChunk]:
        if self.sheet_index != len(dataframe_duplicate_code_chunk.post_state.dfs) - 1:
            return None
        
        return DataframeDuplicateCodeChunk(
            dataframe_duplicate_code_chunk.prev_state,
            self.post_state,
            dataframe_duplicate_code_chunk.sheet_index
        )

    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, DataframeRenameCodeChunk):
            return self._combine_left_with_dataframe_rename(other_code_chunk)
        elif isinstance(other_code_chunk, PivotCodeChunk):
            return self._combine_left_with_pivot_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, MergeCodeChunk):
            return self._combine_left_with_merge_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, SimpleImportCodeChunk) or isinstance(other_code_chunk, ExcelImportCodeChunk) \
            or isinstance(other_code_chunk, ExcelRangeImportCodeChunk) or isinstance(other_code_chunk, SnowflakeImportCodeChunk):
            return self._combine_left_with_import_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, DataframeDuplicateCodeChunk):
            return self._combine_left_with_dataframe_duplicate(other_code_chunk)

        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
from typing import Any, Dict

import pandas as pd
from mitosheet.types import StepsManagerType
from mitosheet.utils import get_row_data_array

# The maximum number of values the front-end sends to the backend
# See comments in function description below.
MAX_UNIQUE_VALUES = 1_000

def get_unique_value_counts(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[str, Any]:
    """
    Sends back a string that can be parsed to a JSON object that
    contains the normalized value counts for the series at column_id 
    in the df at sheet_index.

    Also takes a search_string and sort string, which it uses to filter 
    down the dataset if there are more than MAX_UNIQUE_VALUES
    
    NOTE: the front-end also filters with the search string, and
    reorders with the sort; we just do it here on the backend in 
    the case that there are more than MAX_UNIQUE_VALUES so we 
    don't crash the front-end with too much data.
    """
    sheet_index = params['sheet_index']
    column_id = params['column_id']
    search_string = params['search_string']
    sort = params['sort']

    column_header = steps_manager.curr_step.column_ids.get_column_header_by_id(sheet_index, column_id)
    
    series: pd.Series = steps_manager.dfs[sheet_index][column_header]

    unique_value_counts_percents_series = series.value_counts(normalize=True, dropna=False)
    unique_value_counts_series = series.value_counts(dropna=False)
    
    unique_value_counts_df = pd.DataFrame({
        'values': unique_value_counts_percents_series.index,
        'percents': unique_value_counts_percents_series, 
        'counts': unique_value_counts_series
    })

    if len(unique_value_counts_df) > MAX_UNIQUE_VALUES:
        # First, we turn the series into a string series, so that we can
        # easily filter on it without issues (and sort in some cases)
        new_unique_value_counts_df = unique_value_counts_df.copy(deep=True)
        new_unique_value_counts_df['values_strings'] = new_unique_value_counts_df['values'].astype('str')

        # First, we sort in the order they want
        try:
            if sort == 'Ascending Value':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='values', ascending=True, na_position='first')
            elif sort == 'Descending Value':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='values', ascending=False, na_position='first')
            elif sort == 'Ascending Occurence':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='counts', ascending=True, na_position='first')
            elif sort == 'Descending Occurence':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='counts', ascending=False, na_position='first')
        except:
            # If the sort values throws an exception, then this must be because we have a mixed value type, and so we instead
            # sort on the string representation of the values (as this will always work)
            if sort == 'Ascending Value':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='values_strings', ascending=True, na_position='first')
            elif sort == 'Descending Value':
                new_unique_value_counts_df = new_unique_value_counts_df.sort_values(by='values_strings', ascending=False, na_position='first')

        # Then, we filter with the string. Note that we always filter on the string representation
        # because the front-end sends a string
        new_unique_value_counts_df = new_unique_value_counts_df[new_unique_value_counts_df['values_strings'].str.contains(search_string, na=False, case=False)]

        # Finially, we only take the first MAX_UNIQUE_VALUES
        if len(new_unique_value_counts_df) > MAX_UNIQUE_VALUES:
            new_unique_value_counts_df = new_unique_value_counts_df.head(MAX_UNIQUE_VALUES)
            is_all_data = False
        else:
            is_all_data = True

        # And then we filter the unique values down to these specific values
        unique_value_counts_df = unique_value_counts_df.loc[new_unique_value_counts_df.index]
        unique_value_counts_df.reset_index(drop=True)

    else:
        is_all_data = True
    
    return {
        'uniqueValueRowDataArray': get_row_data_array(unique_value_counts_df),
        'isAllData': is_all_data
    }




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State


class EmptyCodeChunk(CodeChunk):
    """
    A code chunk that generates no code, but has a title and a description
    comment, which is necessary for describing some steps in the step list
    even if they don't have generated code (like graphing).

    Notably, as they don't have generated code, we can still optimize out
    these steps in the code optimization process.
    """

    def __init__(self, prev_state: State, post_state: State, display_name: str, description_comment: str):
        super().__init__(prev_state, post_state)
        self.display_name = display_name
        self.description_comment = description_comment

    def get_display_name(self) -> str:
        return self.display_name
    
    def get_description_comment(self) -> str:
        return self.description_comment

    def get_code(self) -> Tuple[List[str], List[str]]:
        return [], []

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        # We just return the other code chunk, while also updating the prev_state. To avoid
        # causing issues by modifying data, we make a copy of this object
        new_other_code_chunk = deepcopy(other_code_chunk)
        new_other_code_chunk.prev_state = self.prev_state
        return new_other_code_chunk


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
After reading in the arguments passed to the frontend,
this update figures out which of them were dataframes
and which of them were file paths, and updates the 
df names in the steps properly.
"""

from copy import deepcopy
from typing import List
from mitosheet.types import CodeOptions, StepsManagerType
from mitosheet.utils import get_valid_python_identifier

CODE_OPTIONS_UPDATE_EVENT = 'code_options_update'
CODE_OPTIONS_UPDATE_PARAMS = ['code_options']

def execute_args_update(
        steps_manager: StepsManagerType,
        code_options: CodeOptions
    ) -> None:

    # Get the valid function names
    valid_function_name = get_valid_python_identifier(code_options['function_name'], 'function', 'func_')
    final_code_options = deepcopy(code_options)
    final_code_options['function_name'] = valid_function_name

    # Get valid parameter names, and make sure they are unique
    valid_parameter_names: List[str] = []
    for parameter_name, parameter_value in code_options['function_params'].items():
        valid_parameter_name = get_valid_python_identifier(parameter_name, 'parameter', 'param_')
        if valid_parameter_name in valid_parameter_names:
            i = 1
            while valid_parameter_name in valid_parameter_names:
                valid_parameter_name = f"{get_valid_python_identifier(parameter_name, 'parameter', 'param_')}_{i}"
                i += 1

        del final_code_options['function_params'][parameter_name]
        final_code_options['function_params'][valid_parameter_name] = parameter_value
        valid_parameter_names.append(valid_parameter_name)

    steps_manager.code_options = final_code_options

CODE_OPTIONS_UPDATE = {
    'event_type': CODE_OPTIONS_UPDATE_EVENT,
    'params': CODE_OPTIONS_UPDATE_PARAMS,
    'execute': execute_args_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on
numbers.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
from datetime import datetime
import decimal
import math
from typing import Optional, Union
import sys
from mitosheet.errors import MitoError
from mitosheet.is_type_utils import is_bool_dtype, is_datetime_dtype, is_int_dtype
import numpy as np

import pandas as pd

from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.sheet_functions.utils import get_final_result_series_or_primitive, get_index_from_series, get_series_from_primitive_or_series
from mitosheet.public.v3.types.decorators import cast_values_in_all_args_to_type, cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import DatetimeFunctionReturnType, DatetimeRestrictedInputType, FloatFunctonReturnType, IntFunctionReturnType, IntRestrictedInputType, NumberFunctionReturnType, NumberInputType, NumberRestrictedInputType

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def ABS(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "ABS",
        "description": "Returns the absolute value of the passed number or series.",
        "search_terms": ["abs", "absolute value"],
        "examples": [
            "ABS(-1.3)",
            "ABS(A)"
        ],
        "syntax": "ABS(value)",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to take the absolute value of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return abs(arg)

    return arg.abs() # type: ignore


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def AVG(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "AVG",
        "description": "Returns the numerical mean value of the passed numbers and series.",
        "search_terms": ["avg", "average", "mean"],
        "examples": [
            "AVG(1, 2)",
            "AVG(A, B)",
            "AVG(A, 2)"
        ],
        "syntax": "AVG(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or series to consider when calculating the average."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or series to consider when calculating the average."
            }
        ]
    }
    """

    # Calculate the sum using the SUM function
    sum_for_avg = SUM(*argv)

    # Then, we count the number of entries in the arguements. To match Excel's behavior, we do not count 
    # nan values, and we do not count when a RollingRange runs off the end of the sum
    num_entries: Union[pd.Series,int] = 0

    for arg in argv:
        if isinstance(arg, pd.DataFrame):
            num_non_null_values = arg.count().sum()
            num_entries += int(num_non_null_values)

        elif isinstance(arg, RollingRange):
            num_non_null_values_series = arg.apply(lambda df: df.count().sum())
            num_entries += num_non_null_values_series
            
        elif isinstance(arg, pd.Series):
            # Because series are summed row-wise, we need to count the number of non-null
            # values in the row, which we do with this little type cast
            num_entries += (~arg.isna()).astype(int)
            
        elif arg is not None:
            num_entries += 1

    there_are_entries = (isinstance(num_entries, int) and num_entries != 0) or isinstance(num_entries, pd.Series)

    return sum_for_avg / num_entries if there_are_entries else 0

@cast_values_in_arg_to_type('s1', 'number')
@cast_values_in_arg_to_type('s2', 'number')
@handle_sheet_function_errors
def CORR(s1: NumberRestrictedInputType, s2: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "CORR",
        "description": "Computes the correlation between two series, excluding missing values.",
        "search_terms": ["corr", "correlation", "r^2"],
        "examples": [
            "=CORR(A, B)",
            "=CORR(B, A)"
        ],
        "syntax": "CORR(series_one, series_two)",
        "syntax_elements": [{
                "element": "series_one",
                "description": "The number series to convert to calculate the correlation."
            }, {
                "element": "series_two",
                "description": "The number series to convert to calculate the correlation."
            }
        ]
    }
    """
    if isinstance(s1, int) or isinstance(s1, float) or isinstance(s2, int) or isinstance(s2, float):
        return 0
    
    return s1.corr(s2, method='pearson') # type: ignore
 

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def EXP(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "EXP",
        "description": "Returns e, the base of the natural logarithm, raised to the power of passed series.",
        "search_terms": ["exp", "exponent", "log", "natural log"],
        "examples": [
            "=EXP(data)",
            "=EXP(A)"
        ],
        "syntax": "EXP(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to raise e to."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return math.exp(arg)
    
    return pd.Series(np.exp(arg))


@cast_values_in_arg_to_type('arg', 'float')
@handle_sheet_function_errors
def FLOAT(arg: NumberRestrictedInputType) -> FloatFunctonReturnType:
    """
    {
        "function": "FLOAT",
        "description": "Converts a string series to a float series. Any values that fail to convert will return NaN.",
        "search_terms": ["number", "to number"],
        "examples": [
            "=FLOAT(Prices_string)",
            "=FLOAT('123.123')"
        ],
        "syntax": "FLOAT(string_series)",
        "syntax_elements": [{
                "element": "string_series",
                "description": "The series or string to convert to a float."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return float(arg)
    
    return arg.fillna(np.nan) # type: ignore

@cast_values_in_arg_to_type('arg', 'int')
@handle_sheet_function_errors
def INT(arg: IntRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "INT",
        "description": "Converts a string series to a int series. Any values that fail to convert will return 0.",
        "search_terms": ["number", "to integer"],
        "examples": [
            "=INT(Prices_string)",
            "=INT('123')"
        ],
        "syntax": "INT(string_series)",
        "syntax_elements": [{
                "element": "string_series",
                "description": "The series or string to convert to a int."
            }
        ]
    }
    """
    return arg
    


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def KURT(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "KURT",
        "description": "Computes the unbiased kurtosis, a measure of tailedness, of a series, excluding missing values.",
        "search_terms": ["kurtosis"],
        "examples": [
            "=KURT(A)",
            "=KURT(A * B)"
        ],
        "syntax": "KURT(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the unbiased kurtosis of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
        
    return arg.kurt() # type: ignore


@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('base', 'number')
@handle_sheet_function_errors
def LOG(arg: NumberRestrictedInputType, base: Optional[NumberRestrictedInputType]=None) -> NumberFunctionReturnType:
    """
    {
        "function": "LOG",
        "description": "Calculates the logarithm of the passed series with an optional base.",
        "search_terms": ["log", "logarithm", "natural log"],
        "examples": [
            "LOG(10) = 1",
            "LOG(100, 10) = 2"
        ],
        "syntax": "LOG(series, [base])",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to take the logarithm of."
            },
            {
                "element": "base [OPTIONAL]",
                "description": "The base of the logarithm to use. Defaults to 10 if no base is passed."
            }
        ]
    }
    """
    
    if base is None:
        base = 10

    if (isinstance(arg, int) or isinstance(arg, float)) and (isinstance(base, int) or isinstance(base, float)):
        return math.log(arg, base)
    
    index = get_index_from_series(arg, base)
    arg = get_series_from_primitive_or_series(arg, index)
    base = get_series_from_primitive_or_series(base, index).fillna(10)
    
    # See here: https://stackoverflow.com/questions/25169297/numpy-logarithm-with-base-n
    return pd.Series(np.log(arg) / np.log(base)) # type: ignore

@cast_values_in_all_args_to_type('number', ['datetime'])
@handle_sheet_function_errors
def MAX(*argv: Union[NumberInputType, None, DatetimeRestrictedInputType]) -> Union[NumberFunctionReturnType, DatetimeFunctionReturnType]:
    """
    {
        "function": "MAX",
        "description": "Returns the maximum value among the passed arguments.",
        "search_terms": ["max", "maximum", "minimum"],
        "examples": [
            "MAX(10, 11)",
            "MAX(Old_Data, New_Data)"
        ],
        "syntax": "MAX(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the maximum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the maximum value from."
            }
        ]
    }
    """

    # If the user passes numbers, we start the default as the smallest possible number
    # but if it's a datetime, we start the default as the earliest possible date
    default_value = -sys.maxsize - 1
    for arg in argv:
        if isinstance(arg, pd.Series) and is_datetime_dtype(str(arg.dtype)):
            default_value = pd.Timestamp.min
            break
        elif isinstance(arg, pd.Timestamp) or isinstance(arg, datetime):
            default_value = pd.Timestamp.min
            break
        elif isinstance(arg, pd.DataFrame):
            # Check if any series are datetimes
            for col in arg.columns:
                if is_datetime_dtype(str(arg[col].dtype)):
                    default_value = pd.Timestamp.min
                    break
        elif isinstance(arg, RollingRange):
            # Check if any series are datetimes
            for col in arg.obj.columns:
                if is_datetime_dtype(str(arg.obj[col].dtype)):
                    default_value = pd.Timestamp.min
                    break

    result = get_final_result_series_or_primitive(
        default_value,
        argv,
        lambda df: df.max().max(),
        lambda previous_value, new_value: max(previous_value, new_value),
        lambda previous_series, new_series: pd.concat([previous_series, new_series], axis=1).max(axis=1)
    )

    # If we don't find any arguements, we default to 0 -- like Excel -- even for numbers
    kept_default_max_value = not isinstance(result, pd.Series) and result == (default_value)
    return result if not kept_default_max_value else 0


@cast_values_in_all_args_to_type('number', ['datetime'])
@handle_sheet_function_errors
def MIN(*argv: Union[NumberInputType, None, DatetimeRestrictedInputType]) -> Union[NumberFunctionReturnType, DatetimeFunctionReturnType]:
    """
    {
        "function": "MIN",
        "description": "Returns the minimum value among the passed arguments.",
        "search_terms": ["min", "minimum", "maximum"],
        "examples": [
            "MIN(10, 11)",
            "MIN(Old_Data, New_Data)"
        ],
        "syntax": "MIN(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the minumum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the minumum value from."
            }
        ]
    }
    """
    default_value = sys.maxsize
    for arg in argv:
        if isinstance(arg, pd.Series) and is_datetime_dtype(str(arg.dtype)):
            default_value = pd.Timestamp.max
            break
        elif isinstance(arg, pd.Timestamp) or isinstance(arg, datetime):
            default_value = pd.Timestamp.max
            break
        elif isinstance(arg, pd.DataFrame):
            # Check if any series are datetimes
            for col in arg.columns:
                if is_datetime_dtype(str(arg[col].dtype)):
                    default_value = pd.Timestamp.max
                    break
        elif isinstance(arg, RollingRange):
            # Check if any series are datetimes
            for col in arg.obj.columns:
                if is_datetime_dtype(str(arg.obj[col].dtype)):
                    default_value = pd.Timestamp.max
                    break

    result = get_final_result_series_or_primitive(
        default_value,
        argv,
        lambda df: df.min().min(),
        lambda previous_value, new_value: min(previous_value, new_value),
        lambda previous_series, new_series: pd.concat([previous_series, new_series], axis=1).min(axis=1)
    )

    # If we don't find any arguements, we default to 0 -- like Excel
    kept_default_min_value = not isinstance(result, pd.Series) and result == default_value
    return result if not kept_default_min_value else 0


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def MULTIPLY(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "MULTIPLY",
        "description": "Returns the product of two numbers.",
        "search_terms": ["mulitply", "product"],
        "examples": [
            "MULTIPLY(2,3)",
            "MULTIPLY(A,3)"
        ],
        "syntax": "MULTIPLY(factor1, [factor2, ...])",
        "syntax_elements": [{
                "element": "factor1",
                "description": "The first number to multiply."
            },
            {
                "element": "factor2, ... [OPTIONAL]",
                "description": "Additional numbers or series to multiply."
            }
        ]
    }
    """

    return get_final_result_series_or_primitive(
        1,
        argv,
        lambda df: df.prod().prod(),
        lambda previous_value, new_value: previous_value * new_value,
        lambda previous_series, new_series: previous_series * new_series
    )


@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('power', 'number')
@handle_sheet_function_errors
def POWER(arg: NumberRestrictedInputType, power: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "POWER",
        "description": "The POWER function can be used to raise a number to a given power.",
        "search_terms": ["power", "raise", "exponent", "square", "cube"],
        "examples": [
            "POWER(4, 1/2)",
            "POWER(Dose, 2)"
        ],
        "syntax": "POWER(value, exponent)",
        "syntax_elements": [{
                "element": "value",
                "description": "Number to raise to a power."
            },
            {
                "element": "exponent",
                "description": "The number to raise value to."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return arg ** power
    
    return arg.pow(power) # type: ignore

@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('decimals', 'number')
@handle_sheet_function_errors
def ROUND(arg: NumberRestrictedInputType, decimals: Optional[IntRestrictedInputType]=None) -> NumberFunctionReturnType:
    """
    {
        "function": "ROUND",
        "description": "Rounds a number to a given number of decimals.",
        "search_terms": ["round", "decimal", "integer"],
        "examples": [
            "ROUND(1.3)",
            "ROUND(A, 2)"
        ],
        "syntax": "ROUND(value, [decimals])",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to round."
            },
            {
                "element": "decimals",
                "description": " The number of decimals to round to. Default is 0."
            }
        ]
    }
    """

    def excel_round(x: Optional[Union[int, float]], decimals: int) -> Union[int, float, None]:
        if x is None or pd.isna(x): 
            return None

        context = decimal.getcontext()
        context.rounding = decimal.ROUND_HALF_UP
        number = decimal.Decimal(x)
        rounded_number = round(number, decimals)
        return type(x)(rounded_number)

    # If no decimals option is passed, round to no decimals
    if decimals is None:
        decimals = 0

    if (isinstance(arg, int) or isinstance(arg, float)) and isinstance(decimals, int):
        return excel_round(arg, decimals) 

    index = get_index_from_series(arg, decimals)
    arg = get_series_from_primitive_or_series(arg, index).fillna(np.nan)
    decimals = get_series_from_primitive_or_series(decimals, index).fillna(0)

    return pd.Series(
        [excel_round(num, decimal) for num, decimal in zip(arg, decimals)], # type: ignore
        index=index
    )


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def SKEW(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "SKEW",
        "description": "Computes the skew of a series, excluding missing values.",
        "search_terms": [],
        "examples": [
            "=SKEW(A)",
            "=SKEW(A * B)"
        ],
        "syntax": "SKEW(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the skew of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
        
    return arg.skew() # type: ignore

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def STDEV(arg: NumberInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "STDEV",
        "description": "Computes the standard deviation of a series, excluding missing values.",
        "search_terms": ["standard", "deviation", "standard", "distribution"],
        "examples": [
            "=STDEV(A)",
            "=STDEV(A * B)"
        ],
        "syntax": "STDEV(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the standard deviation of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
    elif isinstance(arg, pd.Series):
        return arg.std() # type: ignore
    elif isinstance(arg, pd.DataFrame):
        return arg.stack().std() # We have to compute them all together
    else:
        return arg.apply(lambda x: x.stack().std()) # type: ignore


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def SUM(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "SUM",
        "description": "Returns the sum of the given numbers and series.",
        "search_terms": ["add"],
        "examples": [
            "SUM(10, 11)",
            "SUM(A, B, D, F)",
            "SUM(A, B, D, F)"
        ],
        "syntax": "SUM(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to add together."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to sum."
            }
        ]
    }
    """

    return get_final_result_series_or_primitive(
        0,
        argv,
        lambda df: df.sum().sum(),
        lambda previous_value, new_value: previous_value + new_value,
        lambda previous_series, new_series: previous_series + new_series
    )

@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def SUMPRODUCT(*argv: Union[pd.Series, pd.DataFrame]) -> NumberFunctionReturnType:
    """
    {
        "function": "SUMPRODUCT",
        "description": "Returns the sum of the product of the passed arguments.",
        "search_terms": ["sum product", "sumproduct", "sum", "product", "weighted average"],
        "examples": [
            "SUMPRODUCT(A:A, B:B)",
            "SUMPRODUCT(A:B)"
        ],
        "syntax": "SUMPRODUCT(array1, [array2, ...])",
        "syntax_elements": [{
                "element": "array1",
                "description": "The first array argument whose components you want to multiply and then add."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional series to multiply then add."
            }
        ]
    }
    """

    # We need to make sure that all of the passed arguments are the same length
    # so we can multiply them together. They must be all dataframes or all series
    for arg in argv:
        error = MitoError(
            'invalid_args_error',
            'SUMPRODUCT',
            f"SUMPRODUCT requires all arguments to be the same dimensions.",
            error_modal=False
        )

        length = len(arg)
        if length != len(argv[0]):
            raise error
        
        if type(arg) != type(argv[0]):
            raise error
        
    if isinstance(argv[0], pd.Series):
        # Put all the values into a single dataframe, prod the rows, and then
        # sum the results
        return pd.concat(argv, axis=1).fillna(1).prod(axis=1)
    else:
        # Multiply argv into a single dataframe - making the columns the same first
        result = argv[0]
        for arg in argv[1:]:
            result = result.mul(arg, fill_value=1)

        # And then sum the results
        return result.sum().sum()
        

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def VALUE(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "VALUE",
        "description": "Converts a string series to a number series. Any values that fail to convert will return an NaN.",
        "search_terms": ["number", "to number", "dtype", "convert", "parse"],
        "examples": [
            "=VALUE(A)",
            "=VALUE('123')"
        ],
        "syntax": "VALUE(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to a number."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return arg
    
    return arg.fillna(np.nan) # type: ignore


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def VAR(arg: NumberInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "VAR",
        "description": "Computes the variance of a series, excluding missing values.",
        "search_terms": ["variance"],
        "examples": [
            "=VAR(A)",
            "=VAR(A - B)"
        ],
        "syntax": "VAR(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the variance of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
    elif isinstance(arg, pd.Series):
        return arg.var() # type: ignore
    elif isinstance(arg, pd.DataFrame):
        return arg.stack().var() # type: ignore
    else:
        return arg.apply(lambda x: x.stack().var()) # type: ignore


NUMBER_FUNCTIONS = {
    'ABS': ABS,
    'AVG': AVG,
    'CORR': CORR,
    'FLOAT': FLOAT,
    'INT': INT,
    'EXP': EXP,
    'KURT': KURT,
    'LOG': LOG,
    'MAX': MAX,
    'MIN': MIN,
    'MULTIPLY': MULTIPLY,
    'POWER': POWER,
    'ROUND': ROUND,
    'SKEW': SKEW,
    'SUM': SUM,
    'SUMPRODUCT': SUMPRODUCT,
    'STDEV': STDEV,
    'VALUE': VALUE,
    'VAR': VAR
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code
from mitosheet.types import ColumnID

class SortCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, sort_direction: str):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_id = column_id
        self.sort_direction = sort_direction

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Sorted a column'
    
    def get_description_comment(self) -> str:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        return f'Sorted {column_header} in {self.sort_direction} order'

    def get_code(self) -> Tuple[List[str], List[str]]:
        from mitosheet.step_performers.sort import SORT_DIRECTION_ASCENDING

        # If there is no sort applied in this step, then bail with no code
        from mitosheet.step_performers.sort import SORT_DIRECTION_NONE
        if self.sort_direction == SORT_DIRECTION_NONE:
            return [], []

        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        transpiled_column_header = column_header_to_transpiled_code(column_header)
        
        na_position_string = 'first' if self.sort_direction == SORT_DIRECTION_ASCENDING else 'last'
        
        return [
            f'{self.df_name} = {self.df_name}.sort_values(by={transpiled_column_header}, ascending={self.sort_direction == SORT_DIRECTION_ASCENDING}, na_position=\'{na_position_string}\')', 
        ], []
    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.dataframe_steps.dataframe_delete_code_chunk import DataframeDeleteCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param


class DataframeDeleteStepPerformer(StepPerformer):
    """
    Deletes a dataframe from everywhere in the step.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_delete'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        sheet_index = params['sheet_index']
        old_dataframe_name = prev_state.df_names[sheet_index]
        params['old_dataframe_name'] = old_dataframe_name
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        
        post_state = prev_state.copy()

        # Execute the delete
        delete_dataframe_from_state(post_state, sheet_index)

        return post_state, {
            'pandas_processing_time': 0 # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DataframeDeleteCodeChunk(
                prev_state, 
                post_state, 
                [params['sheet_index']],
                [params['old_dataframe_name']]
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set() # Redo all of them, as order shifts

def delete_dataframe_from_state(state: State, sheet_index: int) -> None:
    state.column_ids.remove_df(sheet_index)
    state.column_formulas.pop(sheet_index)
    state.column_filters.pop(sheet_index)
    state.df_formats.pop(sheet_index)
    state.dfs.pop(sheet_index)
    state.df_names.pop(sheet_index)
    state.df_sources.pop(sheet_index)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.column_headers_transform_code_chunk import ColumnHeadersTransformCodeChunk
from mitosheet.errors import make_column_exists_error

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID

class ColumnHeadersTransformStepPerformer(StepPerformer):
    """
    Allows you to column headers transform.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'column_headers_transform'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        transformation: Any = get_param(params, 'transformation')


        # We make a new state to modify it
        post_state = prev_state.copy() # TODO: update the deep copies

        pandas_start_time = perf_counter()

        df = post_state.dfs[sheet_index]

        if transformation['type'] == 'uppercase' or transformation['type'] == 'lowercase':
            new_columns = []
            renamed_columns = {}
            for col in df.columns:
                if isinstance(col, str):
                    new_column = col.upper() if transformation['type'] == 'uppercase' else col.lower()
                    new_columns.append(new_column)
                    renamed_columns[col] = new_column
                else:
                    new_columns.append(col)
        elif transformation['type'] == 'replace':
            new_columns = []
            renamed_columns = {}
            for col in df.columns:
                if isinstance(col, str):
                    new_column = col.replace(transformation['old'], transformation['new'])
                    new_columns.append(new_column)
                    renamed_columns[col] = new_column
                else:
                    new_columns.append(col)
        else:
            raise ValueError(f'Unknown transformation type: {transformation["type"]}')

        if len(set(new_columns)) < len(new_columns):
            # Get the first duplicated column in new_columns
            seen: Set[str] = set()
            for new_column_header in new_columns:
                if new_column_header in seen:
                    raise make_column_exists_error(new_column_header)
                else:
                    seen.add(new_column_header)
                    
        df.columns = new_columns

        for old_column_header, new_column_header in renamed_columns.items():
            column_id = post_state.column_ids.get_column_id_by_header(sheet_index, old_column_header)
            post_state.column_ids.set_column_header(sheet_index, column_id, new_column_header)
        
        post_state.dfs[sheet_index] = df
        pandas_processing_time = perf_counter() - pandas_start_time


        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'result': {}
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ColumnHeadersTransformCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'transformation')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains functions that are useful for determining the state of the
current user.
"""
import getpass
import hashlib
import os
from datetime import datetime
import sys
from typing import Optional

import pandas as pd
from mitosheet._version import __version__
from mitosheet.user.db import get_user_field
from mitosheet.user.schemas import (UJ_MITOSHEET_ENTERPRISE,
                                    UJ_MITOSHEET_PRO)


try:
    import mitosheet_helper_pro
    MITOSHEET_HELPER_PRO = True
except ImportError:
    MITOSHEET_HELPER_PRO = False
try:
    import mitosheet_helper_enterprise
    MITOSHEET_HELPER_ENTERPRISE = True
except ImportError:
    MITOSHEET_HELPER_ENTERPRISE = False


def is_running_test() -> bool:
    """
    A helper function that quickly returns if the current code is running 
    inside of a test, which is useful for making sure we don't generate 
    tons of logs.
    """
    # Pytest injects PYTEST_CURRENT_TEST into the current environment when running
    running_pytests = "PYTEST_CURRENT_TEST" in os.environ
    # Github injects CI into the environment when running
    running_ci = 'CI' in os.environ and os.environ['CI'] is not None

    return running_pytests or running_ci

def is_on_kuberentes_mito() -> bool:
    """
    Returns True if the user is on Kuberentes Mito, on staging or on app
    """
    user = getpass.getuser()
    return user == 'jovyan'

def is_enterprise() -> bool:
    """
    Helper function for returning if this is a Mito Enterprise
    users
    """
    is_enterprise = get_user_field(UJ_MITOSHEET_ENTERPRISE)

    # This package overides the user.json
    if MITOSHEET_HELPER_ENTERPRISE:
        return MITOSHEET_HELPER_ENTERPRISE

    return is_enterprise if is_enterprise is not None else False

def is_pro() -> bool:
    """
    Helper function for returning if this is a
    pro deployment of mito
    """

    # This package overides the user.json
    if MITOSHEET_HELPER_PRO:
        return MITOSHEET_HELPER_PRO

    # Check if the config is set
    if os.environ.get('MITO_CONFIG_PRO') is not None:
        from mitosheet.enterprise.mito_config import is_env_variable_set_to_true
        return is_env_variable_set_to_true(os.environ.get('MITO_CONFIG_PRO', ''))

    # If you're on Mito Enterprise, then you get all Mito Pro features
    if is_enterprise():
        return True

    pro = get_user_field(UJ_MITOSHEET_PRO)

    return pro if pro is not None else False



def is_local_deployment() -> bool:
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito()  


def get_pandas_version() -> str:
    """
    Returns the pandas version
    """
    return pd.__version__


def get_python_version() -> int:
    """
    Returns the Python version
    """
    return sys.version_info.minor


def check_pro_acccess_code(access_code: Optional[str]) -> bool:
    """Checks if the passed access code is correct, by hashing it and comparing to the hashed value"""
    return access_code is not None and hashlib.sha256(access_code.encode()).hexdigest() == '761a24dea594a8eafe698acfebb77de90bf0826c9400a2543500ee98929ea132'